# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import time
import subprocess

from jenkinsflow.test.cfg import ApiType
from jenkinsflow.test.framework import api_select
from jenkinsflow.test.framework.logger import log, logt


def _abort(log_file, test_file_name, api_type, fixed_prefix, job_name, sleep_time):
    log(log_file, '\n')
    logt(log_file, "Waiting to abort job:", job_name)
    logt(log_file, "args:", test_file_name, fixed_prefix, job_name, sleep_time)
    time.sleep(sleep_time)
    with api_select.api(test_file_name, api_type, fixed_prefix='jenkinsflow_test__' + fixed_prefix + '__', login=True) as api:
        api.job(job_name, max_fails=0, expect_invocations=0, expect_order=None)
    api.poll()
    api.quick_poll()

    abort_me = api.get_job(api.job_name_prefix + job_name)
    logt(log_file, "Abort job:", abort_me)
    abort_me.stop_all()
    logt(log_file, "Aborted")


if __name__ == '__main__':
    job_name = sys.argv[4]
    with open(job_name, 'a+') as log_file:
        _abort(log_file, sys.argv[1], ApiType[sys.argv[2]], sys.argv[3], job_name, int(sys.argv[5]))


def abort(api, job_name, sleep_time):
    """Call this script as a subprocess"""
    if api.api_type == ApiType.MOCK:
        return

    ff = __file__.replace('.pyc', '.py')
    args = [sys.executable, ff, api.file_name, api.api_type.name, api.func_name.replace('test_', ''), job_name, str(sleep_time)]
    with open(job_name, 'w') as log_file:
        logt(log_file, "Invoking abort subprocess.", args)
    subprocess.Popen(args)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, re, abc, signal
from os.path import join as jp
from collections import OrderedDict
from itertools import chain
from enum import Enum

from .ordered_enum import OrderedEnum
from .jenkins_api import BuildResult, Progress, UnknownJobException


_default_poll_interval = 0.5
_default_report_interval = 5
_default_secret_params = '.*passw.*|.*PASSW.*'
_default_secret_params_re = re.compile(_default_secret_params)

_build_result_failures = (BuildResult.FAILURE, BuildResult.ABORTED)


class Propagation(OrderedEnum):
    # pylint: disable=no-init
    NORMAL = 0
    FAILURE_TO_UNSTABLE = 1
    UNCHECKED = 2


class Checking(OrderedEnum):
    # pylint: disable=no-init
    MUST_CHECK = 0
    HAS_UNCHECKED = 1
    FINISHED = 2


class KillType(Enum):
    NONE = 0
    CURRENT = 1
    ALL = 2

    def __bool__(self):
        return self !=  KillType.NONE

    # Python2 compatibility
    __nonzero__ = __bool__


class JobControlException(Exception):
    def __init__(self, message, propagation=Propagation.NORMAL):
        super().__init__(message)
        self.propagation = propagation


class FlowTimeoutException(JobControlException):
    pass


class FlowScopeException(JobControlException):
    pass


class JobNotIdleException(JobControlException):
    pass


class MessageRedefinedException(JobControlException):
    pass


class JobControlFailException(JobControlException, metaclass=abc.ABCMeta):
    pass


class FailedSingleJobException(JobControlFailException):
    def __init__(self, job, propagation):
        msg = "Failed job: " + repr(job) + ", propagation:" + str(propagation)
        super().__init__(msg, propagation)


class MissingJobsException(JobControlFailException):
    def __init__(self, ex):
        super().__init__(str(ex), propagation=Propagation.NORMAL)


class FailedChildJobException(JobControlFailException):
    def __init__(self, flow_job, failed_child_job, propagation):
        msg = "Failed child job in: " + repr(flow_job) + ", child job:" + repr(failed_child_job) + ", propagation:" + str(propagation)
        super().__init__(msg, propagation)


class FailedChildJobsException(JobControlFailException):
    def __init__(self, flow_job, failed_child_jobs, propagation):
        msg = "Failed child jobs in: " + repr(flow_job) + ", child jobs:" + repr(failed_child_jobs) + ", propagation:" + str(propagation)
        super().__init__(msg, propagation)


class FinalResultException(JobControlFailException):
    def __init__(self, build_result):
        msg = "Flow Unsuccessful: {}".format(build_result)
        super().__init__(msg)
        self.result = build_result


class Killed(Exception):
    pass


class _JobControl(metaclass=abc.ABCMeta):
    def __init__(self, parent_flow, securitytoken, max_tries, propagation, secret_params_re, allow_missing_jobs):
        self.parent_flow = parent_flow
        self.top_flow = parent_flow.top_flow

        self.max_tries = max_tries
        self.total_max_tries = self.max_tries * self.parent_flow.total_max_tries

        self.nesting_level = self.parent_flow.nesting_level + 1
        if self.nesting_level != self.top_flow.current_nesting_level + 1:
            raise FlowScopeException("Flow used out of scope")

        self.securitytoken = securitytoken or self.parent_flow.securitytoken
        self.propagation = propagation
        self.secret_params_re = secret_params_re or self.parent_flow.secret_params_re
        self.allow_missing_jobs = allow_missing_jobs if allow_missing_jobs is not None else self.parent_flow.allow_missing_jobs

        self.checking_status = Checking.MUST_CHECK
        self.result = BuildResult.UNKNOWN
        self.tried_times = 0
        self.total_tried_times = 0
        self.invocation_time = None
        self.msg = None

        self.node_id = self.top_flow.next_node_id
        self.top_flow.next_node_id += 1

    def __enter__(self):
        self.top_flow.current_nesting_level += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.top_flow.current_nesting_level -= 1

    @abc.abstractmethod
    def _invocation_id(self):
        """Must return an identification of invocation to distinguish multiple explicit invocations of the same job."""

    @abc.abstractmethod
    def _prepare_first(self):
        """Must be called before the first invocation of a job"""

    def _prepare_to_invoke(self, reset_tried_times=False):
        """Must be called before each invocation of a job, as opposed to _prepare_first, which is called once in entire run"""
        self.checking_status = Checking.MUST_CHECK if self.propagation != Propagation.UNCHECKED else Checking.HAS_UNCHECKED
        self.result = BuildResult.UNKNOWN
        self.tried_times = 1 if reset_tried_times else self.tried_times + 1
        self.total_tried_times += 1
        self.invocation_time = 0

    def _invocation_message(self, controller_type_name, invocation_repr):
        if self.msg is not None:
            print(self.msg)
        print("\n%s %s (%d/%d,%d/%d): %s" % (
            controller_type_name, self._invocation_id(), self.tried_times, self.max_tries, self.total_tried_times, self.total_max_tries, invocation_repr))

    def _must_invoke_set_invocation_time(self):
        if self.invocation_time:
            return False

        self.invocation_time = self.api.time()
        return True

    @abc.abstractmethod
    def _check(self, report_now):
        """Polled by flow controller until the job reaches state 'successful' or tried_times == parent.max_tries * self.max_tries"""

    def _time_msg(self):
        now = self.api.time()
        return "after: %.3fs/%.3fs" % (now - self.invocation_time, now - self.top_flow.start_time)

    @abc.abstractmethod
    def sequence(self):
        """'compact' representaion of flow/job 'name'"""

    @property
    def indentation(self):
        return self.nesting_level * 3 * ' '

    @property
    def api(self):
        return self.top_flow._api

    @property
    def propagate_result(self):
        if self.result in (BuildResult.SUCCESS, BuildResult.SUPERSEDED) or self.propagation == Propagation.UNCHECKED:
            return BuildResult.SUCCESS
        if self.result == BuildResult.UNSTABLE or self.propagation == Propagation.FAILURE_TO_UNSTABLE:
            return BuildResult.UNSTABLE
        return BuildResult.FAILURE

    @property
    def propagate_checking_status(self):
        if self.propagation == Propagation.UNCHECKED:
            return max(Checking.HAS_UNCHECKED, self.checking_status)
        return self.checking_status

    @property
    def remaining_tries(self):
        return max(self.max_tries - self.tried_times, 0)

    @property
    def remaining_total_tries(self):
        return max(self.total_max_tries - self.total_tried_times, 0)

    def message(self, msg):
        """Define a message that will be printed before the invocation of the job or flow on which it is defined.

        Args:
            msg (object): The message that will be printed.
        """

        if self.msg is not None:
            raise MessageRedefinedException("Existing message: " + repr(self.msg) + ", new message: " + repr(msg))
        self.msg = msg

    def __repr__(self):
        return str(self.sequence())

    @abc.abstractmethod
    def last_jobs_in_flow(self):
        """For json graph calculation"""

    @abc.abstractmethod
    def nodes(self, node_to_id):
        """For json graph calculation"""

    @abc.abstractmethod
    def links(self, prev_jobs, node_to_id):
        """For json graph calculation"""


class _SingleJobInvocation(_JobControl):
    """Represents a single flow-invocation of a Jenkins job.

    Multiple invocations of the same job in a single flow are allowed
    Retries are handled by the same instance of this class, but distinct invocations are handled by different instances.
    """

    def __init__(self, parent_flow, securitytoken, job_name_prefix, max_tries, job_name, params, propagation, secret_params_re, allow_missing_jobs):
        for key, value in params.items():
            # Handle parameters passed as int or bool. Booleans will be lowercased!
            if isinstance(value, (bool, int)):
                params[key] = str(value).lower()
        self.params = params
        super().__init__(parent_flow, securitytoken, max_tries, propagation, secret_params_re, allow_missing_jobs)
        # There is no separate retry for individual jobs, so set self.total_max_tries to the same as parent flow!
        self.total_max_tries = self.parent_flow.total_max_tries
        self.job = None
        self.job_invocation = None
        self.old_build_num = None
        self.name = job_name_prefix + job_name
        self.invocation_number = 0
        self.repr_str = ("unchecked " if self.propagation == Propagation.UNCHECKED else "") + "job: " + repr(self.name)
        self.jenkins_baseurl = None
        self._reported_invoked = False
        self._killed = False
        self._display_params = []
        self._set_display_params()

        print(self.indentation + repr(self))

    def _invocation_id(self):
        return ('Invocation-' + str(self.invocation_number)) if self.invocation_number != 0 else 'Invocation'

    def _prepare_first(self, require_job=False):
        previously_invoked = self.top_flow.jobs.get(self.name)
        if previously_invoked:
            if previously_invoked.invocation_number == 0:
                previously_invoked.invocation_number = 1
            self.invocation_number = previously_invoked.invocation_number + 1
        self.top_flow.jobs[self.name] = self

        try:
            self.job = self.api.get_job(self.name)
        except UnknownJobException as ex:
            # TODO? stack trace
            if require_job or not self.allow_missing_jobs and not self.top_flow.kill:
                self.checking_status = Checking.FINISHED
                self.result = BuildResult.FAILURE
                raise MissingJobsException(ex)
            print(self.indentation + repr(self), "- MISSING JOB")
            super()._prepare_to_invoke(reset_tried_times=False)
            return

        self._prepare_to_invoke()
        _result, progress, _ = self.job.job_status()
        if self.top_flow.require_idle and progress != Progress.IDLE and not self.top_flow.kill:
            # Pylint does not like Enum pylint: disable=no-member
            raise JobNotIdleException(repr(self) + " is in state " + progress.name + ". It must be " + Progress.IDLE.name + '.')

        if not previously_invoked:
            # TODO: Don't poll more than once per job for initial status
            print(self.indentation + self._status_message(progress, self.old_build_num, None, 'latest '))

    def _set_display_params(self):
        first = current = OrderedDict()
        last = OrderedDict()

        display_params = dict((key, (value if not self.secret_params_re.search(key) else '******')) for key, value in self.params.items())
        for name in self.top_flow.params_display_order:
            if name == '*':
                current = last
            if name in display_params:
                current[name] = display_params[name]
                del display_params[name]

        for key, value in chain(first.items(), sorted(display_params.items()), last.items()):
            self._display_params.append((key, repr(value)))

    def _show_job_definition(self):
        if self.job:
            print('Defined', self._invocation_id(), self.job.public_uri + (' - parameters:' if self._display_params else ''))
        else:
            print('Defined', self._invocation_id(), repr(self.name) + " - MISSING JOB")
        for key, value in self._display_params:
            print("    ", key, '=', value)
        if self._display_params:
            print("")

    def __repr__(self):
        if self.invocation_number == 0:
            return self.repr_str
        return self.repr_str + ' ' + self._invocation_id()

    def _invoked_message(self):
        print("Build started:", repr(self.name), self._invocation_id() + ' -' if self.invocation_number else '-', self.job_invocation.console_url())

    def _status_message(self, progress, build_num, queued_why, latest=''):
        if progress == Progress.QUEUED:
            msg = (queued_why) if queued_why else ''
        else:
            msg = latest + "build: " + ('#' + str(build_num) if build_num else str(None))
        return repr(self) + " Status " + progress.name + " - " + msg

    def _prepare_to_invoke(self, reset_tried_times=False):
        super()._prepare_to_invoke(reset_tried_times)
        _, _, self.old_build_num = self.job.job_status()
        self._reported_invoked = False

    def _check(self, report_now):
        if self.job is None:
            self._prepare_first(require_job=True)

        self.job.poll()
        if self._must_invoke_set_invocation_time():
            # Don't re-invoke unchecked jobs that are still running
            if self.propagation != Propagation.UNCHECKED:
                self._invocation_message('Job', self.job.public_uri)
                params = self.params if self.params else None
                self.job_invocation = self.job.invoke(securitytoken=self.securitytoken, build_params=params,
                                                      cause=self.top_flow.cause, description=self.top_flow.description)
            elif not self.job_invocation or self.job_invocation.status()[1] == Progress.IDLE:
                self._invocation_message('Job', self.job.public_uri)
                params = self.params if self.params else None
                self.job_invocation = self.job.invoke(securitytoken=self.securitytoken, build_params=params, cause=self.top_flow.cause,
                                                      description=self.top_flow.description)

        result, progress = self.job_invocation.status()
        if not self._reported_invoked and self.job_invocation.build_number is not None:
            if result != BuildResult.SUPERSEDED:
                self._invoked_message()
            self._reported_invoked = True

        if result == BuildResult.UNKNOWN:
            if report_now:
                build_num = self.job_invocation.build_number if self.job_invocation.build_number else self.old_build_num
                print(self._status_message(progress, build_num, self.job_invocation.queued_why))
            return

        # The job has stopped running
        self.checking_status = Checking.FINISHED
        self.result = result

        # Pylint does not like Enum pylint: disable=no-member
        unchecked = (Propagation.UNCHECKED.name + ' ') if self.propagation == Propagation.UNCHECKED else ''

        inv_id_msg = (' ' + self._invocation_id()) if self.invocation_number else ''
        if result != BuildResult.SUPERSEDED:
            print(self, "stopped running")
            print(self._status_message(progress, self.job_invocation.build_number, self.job_invocation.queued_why))
            # Pylint does not like Enum pylint: disable=maybe-no-member
            print(unchecked + self.result.name + ":", repr(self.job.name) + inv_id_msg, "- build:", self.job_invocation.console_url(), self._time_msg())

            if self.result in _build_result_failures:
                raise FailedSingleJobException(self.job, self.propagation)
            return

        # Pylint does not like Enum pylint: disable=maybe-no-member
        print(unchecked + self.result.name + ":", repr(self.job.name) + inv_id_msg)

    def _kill_check(self, report_now, dequeue):
        if self.job is None:
            print(self, "no job")
            self.checking_status = Checking.FINISHED
            return

        self.job.poll()
        if not self._killed:
            self._killed = not dequeue
            if self.top_flow.kill == KillType.ALL:
                if not dequeue:
                    print("Killing all running builds for:", repr(self.name))
                self.job.stop_all()
            elif self.job_invocation:
                if not dequeue:
                    print("Killing build:", repr(self.name), '-', self.job_invocation.console_url())
                self.job_invocation.stop(dequeue)
            else:
                print("Not invoked:", repr(self.name))

        if self.top_flow.kill == KillType.ALL:
            self.result, progress, current_build_num = self.job.job_status()
            if progress != Progress.IDLE and self.old_build_num == current_build_num:
                if report_now and not dequeue:
                    print(self._status_message(progress, None, self.job.queued_why))
                return
        elif self.job_invocation:
            self.result, progress = self.job_invocation.status()
            if progress != Progress.IDLE:
                if report_now and not dequeue:
                    print(self._status_message(progress, self.job_invocation.build_number, self.job_invocation.queued_why))
                return

        # The job has stopped running
        self.checking_status = Checking.FINISHED
        if self.top_flow.kill == KillType.ALL:
            if self.old_build_num == current_build_num:
                print(self, "stopped running")
            else:
                print(self, "was invoked again after kill")
            print(self._status_message(progress, current_build_num, self.job.queued_why))
        elif self.job_invocation:
            print(self, "stopped running")
            print(self._status_message(progress, self.job_invocation.build_number, self.job_invocation.queued_why))

        # Pylint does not like Enum pylint: disable=maybe-no-member
        print(self.result.name + ":", repr(self.job.name))

    def sequence(self):
        return self.name

    def _final_status(self):
        if self.job is not None:
            # Pylint does not like Enum pylint: disable=maybe-no-member
            if self.result in (BuildResult.SUCCESS, BuildResult.SUPERSEDED) and not self.top_flow.kill:
                print(self.indentation + repr(self), self.result.name)
                return

            self.job.poll()
            if self.job_invocation:
                result, progress = self.job_invocation.status()
            else:
                if self.top_flow.kill == KillType.ALL:
                    result, progress, _ = self.job.job_status()
                else:
                    result, progress = BuildResult.UNKNOWN, Progress.IDLE
            progress_msg = ""
            if progress != Progress.IDLE or result == BuildResult.UNKNOWN or self.top_flow.kill:
                progress_msg = "- " + progress.name
            console_url = ""
            if self.result not in (BuildResult.UNKNOWN, BuildResult.DEQUEUED) and not (self.top_flow.kill == KillType.ALL) and self.job_invocation:
                console_url = self.job_invocation.console_url()
            assert isinstance(result, BuildResult)
            print(self.indentation + repr(self), result.name, progress_msg, console_url)
            return

        print(self.indentation + repr(self), "- MISSING JOB")

    def last_jobs_in_flow(self):
        return [self] if self.propagation != Propagation.UNCHECKED else []

    def nodes(self, node_to_id):
        node_name = self.name[self.top_flow.json_strip_index:]
        url = self.job.public_uri if self.job is not None else None

        # For performance reasons use abbreviations
        return [
            OrderedDict(
                (
                    ("id", node_to_id(self)),
                    ("name", node_name),
                    ("url", url),
                    ("tr", [self.max_tries, self.tried_times, self.total_max_tries, self.total_tried_times]),
                    ("nl", self.nesting_level),
                    ("pr", self.propagation.name),
                    # Pylint does not like Enum pylint: disable=maybe-no-member
                    ("cs", self.checking_status.name),
                    ("res", self.result.name),
                    ("it", self.invocation_time),
                    ("params", self._display_params),
                )
            )
        ]

    def links(self, prev_jobs, node_to_id):
        return [OrderedDict((("source", node_to_id(job)), ("target", node_to_id(self)))) for job in prev_jobs]


# Retries are handled in the _Flow classes instead of _SingleJobInvocation since the individual jobs don't know
# how to retry. The _Serial flow is retried from start of flow and in _Parallel flow individual jobs
# are retried immediately

class _Flow(_JobControl, metaclass=abc.ABCMeta):
    _enter_str = None
    _exit_str = None

    def __init__(self, parent_flow, timeout, securitytoken, job_name_prefix, max_tries, propagation, report_interval, secret_params, allow_missing_jobs):
        secret_params_re = re.compile(secret_params) if isinstance(secret_params, str) else secret_params
        super().__init__(parent_flow, securitytoken, max_tries, propagation, secret_params_re, allow_missing_jobs)
        self.timeout = timeout
        self.job_name_prefix = self.parent_flow.job_name_prefix + job_name_prefix if job_name_prefix is not None else ""
        self.report_interval = report_interval or self.parent_flow.report_interval

        self.invocations = []
        self.last_report_time = 0
        self.last_json_time = 0
        self._failed_child_jobs = {}
        self._can_raise_kill = False

    def parallel(self, timeout=0, securitytoken=None, job_name_prefix='', max_tries=1, propagation=Propagation.NORMAL, report_interval=None, secret_params=None, allow_missing_jobs=None):
        """Defines a parallel flow where nested jobs or flows are executed simultaneously.

        Only differences to :py:meth:`.serial` are described.

        Args:
            max_tries (int): Maximum number of times to invoke the jobs in the flow. Default is 1, meaning no retry will be attempted in case jobs fails.
                If a job fails it is immediately retried::

                    with parallel(..., max_tries=3) as sf:
                        sf.invoke('a', ...)
                        sf.invoke('b', ...)  # fail -> restart job 'b'
                        sf.invoke('c', ...)

        Returns:
            parallel flow object
        """

        assert isinstance(propagation, Propagation)
        return _Parallel(self, timeout, securitytoken, job_name_prefix, max_tries, propagation, report_interval, secret_params, allow_missing_jobs)

    def serial(self, timeout=0, securitytoken=None, job_name_prefix='', max_tries=1, propagation=Propagation.NORMAL, report_interval=None, secret_params=None, allow_missing_jobs=None):
        """Defines a serial flow where nested jobs or flows are executed in order.

        Args:
            timeout (float): Maximum time in seconds to wait for flow jobs to finish. 0 means infinite, however, this flow can not run longer than
                the minimum timeout of any parent flows. Note that jenkins jobs are NOT terminated when the flow times out.

            securitytoken (str): Token to use on security enabled Jenkins instead of username/password. The Jenkins job must have the token configured.
                If None, the parent flow securitytoken is used.

            job_name_prefix (str): All jobs defined in flow will automatically be prefixed with the parent flow job_name_prefix + this job_name_prefix
                before invoking Jenkins job. To reset prefixing (i.e. don't use parent flow prefix either), set the value to None

            max_tries (int): Maximum number of times to invoke the flow. Default is 1, meaning no retry will be attempted in case a job fails.
                If a job fails, jobs are retried from start of the parallel flow::

                    with serial(..., max_tries=3) as sf:
                        sf.invoke('a', ...)
                        sf.invoke('b', ...)  # fail -> restart flow from job 'a'
                        sf.invoke('c', ...)

                Retries may be nested::

                    with parallel(..., max_tries=2) as pf:
                        with pf.serial(..., max_tries=3) as sf:
                            sf.invoke('a', ...)  # If job 'a' fails it could be invoked up to 6 times

            propagation (Propagation): How to propagate errors from failed Jenkins jobs.
                This can be used to downgrade a 'FAILURE' result to 'UNSTABLE'.
                This will not change the result of a failed job, the result of jobs are used in caculating the propagated result. E.g::

                    with parallel(...) as pf:
                        with pf.serial(..., propagation=Propagation.UNSTABLE) as sf1:
                            sf1.invoke('a', ...)  # If job 'a' fails the result propagated to sf1 will be 'UNSTABLE' and sf1 will propagate
                                                  #  'UNSTABLE' on to pf

                        with pf.serial(...) as sf2:
                            sf2.invoke('b', ...)  # If job 'b' fails the result propagated to sf2 and pf will be 'FAILURE'

                    sys.exit(77 if pf.result == Propagation.UNSTABLE else 0)
                    # Assuming the Jenkins job was configured with 'Exit code to set build unstable==77'

                NOTE: You must make sure that the jobs are configured correctly and that the correct exit code is used in the shell step, otherwise
                the propagation value will have no effect on the final job status.
                NOTE: Also see the raise_if_unsuccessful argument to the top level `serial` and `parallel` flows.

            report_interval (float): The interval in seconds between reporting the status of polled Jenkins jobs.
                If None the parent flow report_interval is used.

            secret_params (re.RegexObject): Regex of Jenkins job invocation parameter names, for which the value will be masked out with '******' when
                parameters are printed.
                If None the parent flow secret_params is used.

            allow_missing_jobs (boolean): If true it is not considered an error if Jenkins jobs are missing when the flow starts.
                It is assumed that the missing jobs are created by jobs in the flow, prior to the missing jobs being invoked.
                If None the parent flow allow_missing_jobs is used.

        Returns:
            serial flow object

        Raises:
            JobControlException
        """

        assert isinstance(propagation, Propagation)
        return _Serial(self, timeout, securitytoken, job_name_prefix, max_tries, propagation, report_interval, secret_params, allow_missing_jobs)

    def invoke(self, job_name, **params):
        """Define a Jenkins job invocation that will be invoked under control of the surrounding flow.

        This does not create the job in Jenkins. It defines how the job will be invoked by ``jenkinsflow``.

        Args:
            job_name (str): The last part of the name of the job in jenkins.
                If the surrounding flow sets the :py:obj:`job_name_prefix` the actual name of the invoked job will be the parent flow job_name_prefix + job_name.
            **params (str, int, boolean): Arguments passed to Jenkins when invoking the job. Strings are passed as they are,
                booleans are automatically converted to strings and lowercased, integers are automatically converted to strings.
        """

        inv = _SingleJobInvocation(self, self.securitytoken, self.job_name_prefix, self.max_tries, job_name, params, self.propagation, self.secret_params_re, self.allow_missing_jobs)
        self.invocations.append(inv)
        return inv

    def invoke_unchecked(self, job_name, **params):
        """Define a Jenkins job invocation that will be invoked under control of the surrounding flow, but will never cause the flow to fail.

        The job is always run in parallel with other jobs in the flow, even when invoked in a serial flow.
        It is not started out of order, but following jobs will be started immediately after this is started.
        The job will be monitored and reported on, only as long as regular "checked" jobs are still running.
        If it fails, it may be retried (depending on surrounding flows max_tries option), but only as long as regular jobs are still running.
        If the job is still running when all normal jobs are finished, the flow will exit, and the job is left running.

        See :py:meth:`invoke` for parameter description.
        """

        inv = _SingleJobInvocation(self, self.securitytoken, self.job_name_prefix, self.max_tries, job_name, params, Propagation.UNCHECKED, self.secret_params_re, self.allow_missing_jobs)
        self.invocations.append(inv)
        return inv

    def _invocation_id(self):
        return 'Invocation'

    def _prepare_first(self):
        print(self.indentation + self._enter_str)
        self._prepare_to_invoke()
        for inv in self.invocations:
            inv._prepare_first()
        print(self.indentation + self._exit_str)

    def _show_job_definition(self):
        for inv in self.invocations:
            inv._show_job_definition()

    def _final_status(self):
        print(self.indentation + self._enter_str)
        for inv in self.invocations:
            inv._final_status()
        print(self.indentation + self._exit_str)

    def _check_timeout(self):
        now = self.api.time()
        if self.timeout and now - self.invocation_time > self.timeout:
            unfinished_msg = ". Unfinished jobs:" + repr([inv.sequence() for inv in self.invocations if inv.checking_status == Checking.MUST_CHECK])
            raise FlowTimeoutException("Timeout " + self._time_msg() + ", in flow " + str(self) + unfinished_msg, self.propagation)

    def __enter__(self):
        print(self.indentation + self._enter_str)
        super().__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None

        print(self.indentation + self._exit_str)
        super().__exit__(exc_type, exc_value, traceback)
        if self.parent_flow:
            # Insert myself in parent if I'm not empty
            if self.invocations:
                self.parent_flow.invocations.append(self)
                return

            print(self.indentation + "INFO: Ignoring empty flow")
        print()

    def _check_report(self):
        now = self.api.time()
        report_now = now - self.last_report_time >= self.report_interval
        if report_now:
            self.last_report_time = now
        return report_now

    def _check_invoke_report(self):
        if self._must_invoke_set_invocation_time():
            self._invocation_message('Flow', self)
        return self._check_report()

    def _kill_check(self, report_now, dequeue):
        report_now = self._check_report()

        checking_status = Checking.FINISHED
        for inv in self.invocations:
            if inv.checking_status != Checking.FINISHED:
                inv._kill_check(report_now, dequeue)
                job_propagate_checking_status = inv.checking_status if inv.checking_status != Checking.HAS_UNCHECKED else Checking.MUST_CHECK
                checking_status = min(checking_status, job_propagate_checking_status)

        self.checking_status = checking_status
        if self.checking_status == Checking.FINISHED:
            # All jobs have stopped running
            for inv in self.invocations:
                self.result = min(self.result, inv.propagate_result)
            self.report_result()

    def report_result(self):
        # Pylint does not like Enum pylint: disable=no-member
        unchecked = (Propagation.UNCHECKED.name + ' ') if self.propagation == Propagation.UNCHECKED else ''
        # Pylint does not like Enum pylint: disable=maybe-no-member
        print('Flow ' + unchecked + self.result.name, self, self._time_msg())

    def json(self, file_path, indent=None):
        node_to_id = lambda job: job.node_id
        separators=None
        if indent:
            node_to_id = lambda job: job.name
            separators=(',', ': ')

        nodes = self.nodes(node_to_id)
        links = self.links([], node_to_id)
        graph = OrderedDict((('nodes', nodes), ('links', links)))

        import json
        from atomicfile import AtomicFile
        if file_path is not None:
            with AtomicFile(file_path, 'w+') as out_file:
                # python3 doesn't need  separators=(',', ': ')
                json.dump(graph, out_file, indent=indent, separators=separators)
        else:
            # python3 doesn't need  separators=(',', ': ')
            return json.dumps(graph, indent=indent, separators=separators)


class _Parallel(_Flow):
    _enter_str = "parallel flow: ("
    _exit_str = ")\n"

    def _check(self, report_now):
        report_now = self._check_invoke_report()

        checking_status = Checking.FINISHED
        for inv in self.invocations:
            try:
                if inv.checking_status != Checking.FINISHED:
                    inv._check(report_now)
                    checking_status = min(checking_status, inv.propagate_checking_status)
                    if id(inv) in self._failed_child_jobs:
                        del self._failed_child_jobs[id(inv)]
            except JobControlFailException:
                self._failed_child_jobs[id(inv)] = inv

                if inv.result == BuildResult.ABORTED:
                    if inv.remaining_tries or inv.remaining_total_tries:
                        print("ABORTED:", inv, "not retrying")
                    inv.checking_status = Checking.FINISHED
                    continue

                if inv.remaining_tries:
                    print("RETRY:", inv, "failed but will be retried. Up to", inv.remaining_tries, "more times in current flow")
                    checking_status = Checking.MUST_CHECK
                    inv._prepare_to_invoke()
                    continue

                if inv.remaining_total_tries:
                    print("RETRY:", inv, "failed but will be retried. Up to", inv.remaining_total_tries, "more times through outer flow")
                    inv._prepare_to_invoke(reset_tried_times=True)
                    continue

                inv.checking_status = Checking.FINISHED

        self.checking_status = checking_status
        if self.checking_status != Checking.MUST_CHECK and self.result == BuildResult.UNKNOWN:
            # All jobs have stopped running or are 'unchecked'
            for inv in self.invocations:
                self.result = min(self.result, inv.propagate_result)
            self.report_result()

            if self.result in _build_result_failures:
                raise FailedChildJobsException(self, self._failed_child_jobs.values(), self.propagation)
        else:
            self._check_timeout()

    def sequence(self):
        return tuple([inv.sequence() for inv in self.invocations])

    def last_jobs_in_flow(self):
        invocations = []
        for inv in self.invocations:
            invocations.extend(inv.last_jobs_in_flow())
        return invocations

    def nodes(self, node_to_id):
        nodes = []
        for inv in self.invocations:
            child_nodes = inv.nodes(node_to_id)
            nodes.extend(child_nodes)
        return nodes

    def links(self, prev_jobs, node_to_id):
        links = []
        for inv in self.invocations:
            child_links = inv.links(prev_jobs, node_to_id)
            links.extend(child_links)
        return links


class _Serial(_Flow):
    _enter_str = "serial flow: ["
    _exit_str = "]\n"

    def __init__(self, parent_flow, securitytoken, timeout, job_name_prefix='', max_tries=1, propagation=Propagation.NORMAL,
                 report_interval=None, secret_params=_default_secret_params_re, allow_missing_jobs=None):
        super().__init__(parent_flow, securitytoken, timeout, job_name_prefix, max_tries, propagation,
                                      report_interval, secret_params, allow_missing_jobs)
        self.job_index = 0

    def _prepare_to_invoke(self, reset_tried_times=False):
        super()._prepare_to_invoke(reset_tried_times)
        self.job_index = 0

    def _check(self, report_now):
        report_now = self._check_invoke_report()

        checking_status = Checking.FINISHED
        for inv in self.invocations[0:self.job_index + 1]:
            try:
                if inv.checking_status != Checking.FINISHED:
                    inv._check(report_now)
                    checking_status = min(checking_status, inv.propagate_checking_status)
            except JobControlFailException:
                if inv.result == BuildResult.ABORTED:
                    if inv.remaining_tries or inv.remaining_total_tries:
                        print("ABORTED:", inv, "not retrying")
                    inv.checking_status = Checking.FINISHED
                    self.total_max_tries = 0
                    self.max_tries = 0
                    continue

                # The job has stopped running
                if inv.remaining_tries:
                    if inv.propagation != Propagation.NORMAL:
                        print("MAY RETRY:", inv, inv.propagation, " failed, will only retry if checked failures. Up to", inv.remaining_tries, "more times in current flow")
                        continue
                    print("RETRY:", inv, "failed, retrying child jobs from beginning. Up to", inv.remaining_tries, "more times in current flow")
                    checking_status = Checking.MUST_CHECK
                    for pre_job in self.invocations[0:self.job_index + 1]:
                        pre_job._prepare_to_invoke()
                    self.job_index = 0
                    continue

                if inv.remaining_total_tries:
                    if inv.propagation != Propagation.NORMAL:
                        print("MAY RETRY:", inv, inv.propagation, " failed, will only retry if checked failures. Up to", inv.remaining_total_tries, "more times through outer flow")
                        continue
                    print("RETRY:", inv, "failed, retrying child jobs from beginning. Up to", inv.remaining_total_tries, "more times through outer flow")
                    for pre_job in self.invocations[0:self.job_index + 1]:
                        pre_job._prepare_to_invoke(reset_tried_times=True)
                    self.job_index = 0
                    continue

                inv.checking_status = Checking.FINISHED
                if inv.propagation != Propagation.UNCHECKED:
                    self.job_index = len(self.invocations)

        self.checking_status = checking_status
        if self.checking_status != Checking.MUST_CHECK and self.result == BuildResult.UNKNOWN:
            for inv in self.invocations[0:self.job_index + 1]:
                self.result = min(self.result, inv.propagate_result)

            if self.result in _build_result_failures:
                self.report_result()
                raise FailedChildJobException(self, inv, self.propagation)

            self.job_index += 1
            if self.job_index < len(self.invocations):
                self.checking_status = Checking.MUST_CHECK
                self.result = BuildResult.UNKNOWN
                return

            # All jobs have stopped running or are 'unchecked'
            self.report_result()
        else:
            self._check_timeout()

    def sequence(self):
        return [inv.sequence() for inv in self.invocations]

    def last_jobs_in_flow(self):
        for inv in self.invocations[-1:0:-1]:
            last_jobs = inv.last_jobs_in_flow()
            if last_jobs:
                return last_jobs
        return []

    def nodes(self, node_to_id):
        nodes = []
        for inv in self.invocations:
            child_nodes = inv.nodes(node_to_id)
            nodes.extend(child_nodes)
        return nodes

    def links(self, prev_jobs, node_to_id):
        links = []
        for inv in self.invocations:
            child_links = inv.links(prev_jobs, node_to_id)
            links.extend(child_links)
            prev_jobs = inv.last_jobs_in_flow()
        return links


class _TopLevelControllerMixin(metaclass=abc.ABCMeta):
    def toplevel_init(self, jenkins_api, securitytoken, username, password, top_level_job_name_prefix, poll_interval, direct_url, require_idle,
                      json_dir, json_indent, json_strip_top_level_prefix, params_display_order, just_dump, kill_all, description, raise_if_unsuccessful):
        self._start_msg()
        # pylint: disable=attribute-defined-outside-init
        # Note: Special handling in top level flow, these atributes will be modified in proper flow init
        self.parent_flow = self
        self.top_flow = self
        self.job_name_prefix = ''
        self.total_max_tries = 1
        self.nesting_level = -1
        self.current_nesting_level = -1
        self.report_interval = _default_report_interval
        self.secret_params_re = _default_secret_params_re
        self.allow_missing_jobs = None
        self.next_node_id = 0
        self.just_dump = just_dump

        self.kill = KillType.ALL if kill_all else KillType.NONE

        jenkins_job_name = os.environ.get('JOB_NAME')
        if jenkins_job_name:
            self.cause = "By flow job " + repr(jenkins_job_name) + ' #' +  os.environ.get('BUILD_NUMBER', ' None')
        else:
            import getpass
            user = getpass.getuser()
            self.cause = "By flow script, user: " + user

        self._api = jenkins_api
        self.securitytoken = securitytoken
        self.username = username or self._api.username
        self.password = password or self._api.password
        self.poll_interval = poll_interval
        self.direct_url = direct_url.rstrip('/') if direct_url is not None else direct_url
        self.require_idle = require_idle

        self.json_dir = json_dir
        self.json_indent = json_indent
        self.json_strip_index = len(top_level_job_name_prefix) if json_strip_top_level_prefix else 0
        self.json_file = jp(self.json_dir, 'flow_graph.json') if json_dir is not None else None

        self.params_display_order = params_display_order
        self.description = description
        self.raise_if_unsuccessful = raise_if_unsuccessful

        # 'jobs' hold unique jobs by name vs 'invocations' on individual flows which may hold multiple invocations on one job
        self.jobs = {}

        # Set signalhandler to kill entire flow
        def set_kill(_sig, _frame):
            print("\nGot SIGTERM: Killing all builds belonging to current flow")
            self.kill = KillType.CURRENT
            if self._can_raise_kill:
                raise Killed()
        signal.signal(signal.SIGTERM, set_kill)

        # Allow test framework to set securitytoken, so that we won't have to litter all the testcases with it
        return self.securitytoken or jenkins_api.securitytoken if hasattr(jenkins_api, 'securitytoken') else None

    @staticmethod
    def _start_msg():
        print()
        print("=== Jenkinsflow ===")
        print()
        print("Legend:")
        print("Serial builds: []")
        print("Parallel builds: ()")
        print("Invocation-N (w/x,y/z): ")
        print("    -N: 'Invocation N of same job', where N is invocation number which is increased every time a job has been explicitly")
        print("            invoked (as opposed to retried). '-N' is only present for jobs with multiple invocations.")
        print("    w=current retry invocation in current flow scope, x=max in scope, y=total number of invocations, z=total max invocations")
        print("Elapsed time: 'after: x/y': x=time spent during current run of job, y=time elapsed since start of outermost flow")
        print()
        print("--- Calculating flow graph ---")

    def wait_for_jobs(self):
        if self.json_file:
            self.json(self.json_file, self.json_indent)

        if self.just_dump:
            return

        if not self.invocations:
            print("WARNING: Empty toplevel flow", self, "nothing to do.")
            return

        # Wait for jobs to finish
        print()
        print("--- Getting initial job status ---")
        self.api.poll()
        self._prepare_first()
        self._show_job_definition()

        if self.json_file:
            self.json(self.json_file, self.json_indent)

        # pylint: disable=attribute-defined-outside-init
        self.start_time = self._api.time()
        self.last_report_time = self.start_time
        last_json_time = self.start_time
        json_interval = max(30, self.report_interval)

        print()
        if not self.kill:
            print("--- Starting flow ---")
        else:
            print("--- Starting kill of all builds in flow ---")

        sleep_time = min(self.poll_interval, self.report_interval)
        try:
            dequeue = True
            while self.checking_status == Checking.MUST_CHECK:
                self.api.quick_poll()
                if not self.kill:
                    try:
                        self._can_raise_kill = True
                        self._check(None)
                        self._can_raise_kill = False
                    except Killed:
                        pass
                else:
                    self.api.queue_poll()
                    self._kill_check(None, dequeue)
                    dequeue = False

                self._api.sleep(sleep_time)
                if self.json_file:
                    now = self._api.time()
                    json_now = now - last_json_time >= json_interval
                    if json_now:
                        last_json_time = now
                    self.json(self.json_file, self.json_indent)

            if self.raise_if_unsuccessful and self.result != BuildResult.SUCCESS and self.kill != KillType.ALL:
                raise FinalResultException(self.result)
        finally:
            print()
            print("--- Final status ---")
            self.api.quick_poll()
            self._final_status()
            if self.json_file:
                self.json(self.json_file, self.json_indent)


class parallel(_Parallel, _TopLevelControllerMixin):
    """Defines a parallel flow where nested jobs or flows are executed simultaneously.

    See :py:class:`serial` and :py:meth:`_Flow.parallel` for a description.
    """

    def __init__(self, jenkins_api, timeout, securitytoken=None, username=None, password=None, job_name_prefix='', max_tries=1, propagation=Propagation.NORMAL,
                 report_interval=_default_report_interval, poll_interval=_default_poll_interval, secret_params=_default_secret_params_re, allow_missing_jobs=False,
                 json_dir=None, json_indent=None, json_strip_top_level_prefix=True, direct_url=None, require_idle=True, just_dump=False, params_display_order=(),
                 kill_all=False, description=None, raise_if_unsuccessful=True):
        assert isinstance(propagation, Propagation)
        securitytoken = self.toplevel_init(jenkins_api, securitytoken, username, password, job_name_prefix, poll_interval, direct_url, require_idle,
                                           json_dir, json_indent, json_strip_top_level_prefix, params_display_order, just_dump, kill_all, description=description,
                                           raise_if_unsuccessful=raise_if_unsuccessful)
        super().__init__(self, timeout, securitytoken, job_name_prefix, max_tries, propagation, report_interval, secret_params, allow_missing_jobs)
        self.parent_flow = None

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None

        super().__exit__(exc_type, exc_value, traceback)
        self.wait_for_jobs()


class serial(_Serial, _TopLevelControllerMixin):
    r"""Defines a serial flow where nested jobs or flows are executed in order.

    Only differences to  :py:meth:`_Flow.serial` are described.

    Args:
        jenkins_api (:py:class:`.jenkins_api.Jenkins` or :py:class:`.script_api.Jenkins`): Jenkins Api instance used for accessing jenkins.
            If jenkins_api is instantiated with username/password you do not need to specify username/password to the flow (see below).

        securitytoken (str): Token to use on security enabled Jenkins instead of username/password. The Jenkins job must have the token configured.

        username (str): Name of user authorized to run Jenkins 'cli' and change job status.

        password (str): Password of user.
            The username/password here is are not used for running the jobs. See jenkins_api for that.
            If username/password is specified for jenkins_api, they will be used unless they are also specified on the flow.

        job_name_prefix (str): All jobs defined in flow will automatically be prefixed with this string before invoking Jenkins job.

        poll_interval (float): The interval in seconds between polling the status of unfinished Jenkins jobs.

        allow_missing_jobs (boolean): If true it is not considered an error if Jenkins jobs are missing when the flow starts.
            It is assumed that the missing jobs are created by other jobs in the flow

        json_dir (str): Directory in which to generate flow graph json file. If None, no flow graph is generated.

        json_indent (int): If not None json graph file is pretty printed with this indentation level.

        json_strip_top_level_prefix (boolean): If True, the job_name_prefix will be stripped from job names when generating json graph file

        direct_url (str): Non proxied url for accessing Jenkins, use this as an optimization to avoid routing rest calls from Jenkins through
            a proxy if the JENKINS_URL setting does not point directly to jenkins.

        require_idle (boolean): If True it is considered an error if any of the jobs in the flow are running when the flow starts.

        just_dump (boolean): If True, the flow is just printed, no jobs are invoked.

        params_display_order (list): List of job parameter names used for ordering the parameters in the output.
            The format is [first1, ..., firstN, '\*', last1, ..., lastN], where first..., last... are names that will be matched against the
            invoke \*\*param names.
            Any of first..., '*', last... may be omitted
            Any parameters that are not matched will be displayes at the place of the '*', if specified, otherwise they will be displayed last.

        kill_all (boolean): If True, all running builds for jobs defined in the flow will be aborted, regardless which flow invocation
            started the build.
            Note: It also possible to send SIGTERM to an already running flow to make the flow abort all builds started by the current
            invocation of the flow, but not builds started by other invocations of the same flow.

        raise_if_unsuccessful (bool): If the result of the outermost flow is not `BuildResult.SUCCESS` and no exception was raised, then
            a `FinalResultException` will be raised. The result property of this exception should be checked and the proper value returned
            from the shell step. Use in combination with the 'Exit code to set build unstable' feature in the advanced section on freestyle
            jobs shell build step. E.g::

                try:
                    with parallel(api) as ctrl1:
                        ctrl1.invoke('j11')

                        with ctrl1.serial() as ctrl2:
                            ctrl2.invoke('j21')  # ends succesfully
                            ctrl2.invoke('j22')  # ends in state UNSTABLE
                            ctrl2.invoke('j23')

                except FinalResultException as ex:
                    if ex.result == BuildResult.UNSTABLE:
                        return 77
                    raise

            If set to False, then the propagated reult value is available as the attribute 'result' on the top level flow, so that it can be used to
            return the proper value.

    Returns:
        serial flow object

    Raises:
        JobControlException
    """

    def __init__(self, jenkins_api, timeout, securitytoken=None, username=None, password=None, job_name_prefix='', max_tries=1, propagation=Propagation.NORMAL,
                 report_interval=_default_report_interval, poll_interval=_default_poll_interval, secret_params=_default_secret_params_re, allow_missing_jobs=False,
                 json_dir=None, json_indent=None, json_strip_top_level_prefix=True, direct_url=None, require_idle=True, just_dump=False, params_display_order=(),
                 kill_all=False, description=None, raise_if_unsuccessful=True):
        assert isinstance(propagation, Propagation)
        securitytoken = self.toplevel_init(jenkins_api, securitytoken, username, password, job_name_prefix, poll_interval, direct_url, require_idle,
                                           json_dir, json_indent, json_strip_top_level_prefix, params_display_order, just_dump, kill_all, description=description,
                                           raise_if_unsuccessful=raise_if_unsuccessful)
        super().__init__(self, timeout, securitytoken, job_name_prefix, max_tries, propagation, report_interval, secret_params, allow_missing_jobs)
        self.parent_flow = None

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None

        super().__exit__(exc_type, exc_value, traceback)
        self.wait_for_jobs()



import os.path

from .api_base import AuthError, ClientError


def _join_url(start, end):
    return os.path.join(start.rstrip('/'), end.lstrip('/'))


class ResourceNotFound(Exception):
    pass


class RequestsRestApi():
    def __init__(self, direct_uri, username, password):
        super().__init__()
        import requests
        self.session = requests.Session()
        if username or password:
            self.session.auth = requests.auth.HTTPBasicAuth(username, password.encode("utf-8"))
        self.direct_uri = direct_uri

    @staticmethod
    def _check_response(response, good_responses=(200,)):
        if response.status_code in good_responses:
            return response

        try:
            response.raise_for_status()
        except Exception as ex:
            if response.status_code == 404:
                raise ResourceNotFound(ex)
            if response.status_code == 401:
                raise AuthError(ex)
            if response.status_code == 403:
                raise ClientError(ex)
            raise

        # TODO: This is dubious, maybe we should raise here instead.
        return response

    def _get(self, url, params):
        import requests
        try:
            return self._check_response(self.session.get(_join_url(self.direct_uri, url), params=params))
        except requests.ConnectionError as ex:
            raise ConnectionError(ex)

    def get_content(self, url, **params):
        return self._get(url, params=params).content

    def get_json(self, url="", **params):
        return self._get(_join_url(url, "api/json"), params=params).json()

    def post(self, url, payload=None, headers=None, allow_redirects=False, **params):
        response = self.session.post(_join_url(self.direct_uri, url), headers=headers, data=payload, allow_redirects=allow_redirects, params=params)
        return self._check_response(response, (200, 201))

    def headers(self, allow_redirects=True):
        return self._check_response(self.session.head(self.direct_uri, allow_redirects=allow_redirects)).headers


def _check_restkit_response(func):
    def deco(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except self.restkit.Unauthorized as ex:
            raise AuthError(ex.response.status + " user: '" + self.username + "' for url: " + ex.response.final_url)
        except self.restkit.errors.ResourceNotFound as ex:
            raise ResourceNotFound(str(ex))

    return deco


class RestkitRestApi():
    """Note: This is no longer supported or tested, but may still work."""
    def __init__(self, direct_uri, username, password, **kwargs):
        super().__init__()

        import restkit
        self.restkit = restkit
        import json
        self.json = json

        if username or password:
            filters = kwargs.get('filters', [])
            filters.append(restkit.BasicAuth(username, password))
            kwargs['filters'] = filters

        self.username = username
        self.session = restkit.Resource(uri=direct_uri, **kwargs)

    @_check_restkit_response
    def get_content(self, url, **params):
        return self.session.get(url, **params).body_string()

    @_check_restkit_response
    def get_json(self, path="", headers=None, params_dict=None, **params):
        response = self.session.get(path=_join_url(path, "api/json"), headers=headers, params_dict=params_dict, **params)
        return self.json.loads(response.body_string())

    @_check_restkit_response
    def post(self, url, payload=None, headers=None, allow_redirects=False, **params):  # pylint: disable=unused-argument
        return self.session.post(url, headers=headers, payload=payload, **params)

    @_check_restkit_response
    def headers(self, allow_redirects=True):  # pylint: disable=unused-argument
        return self.session.head().headers



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, os

import click

# Allow relative imports while running as script. Necessary for testing without installing
if __package__ is None:
    _here = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(1, os.path.dirname(os.path.dirname(_here)))
    import jenkinsflow.cli
    __package__ = "jenkinsflow.cli"


from .set_build_description import set_build_description


@click.group()
def cli():
    """Commandline utilities for jenkinsflow"""
    pass


cli.add_command(set_build_description, name="set_build_description")


if __name__ == "__main__":
    cli(auto_envvar_prefix='JENKINSFLOW')



import os, socket
from enum import Enum

env_var_prefix = "JENKINSFLOW_"

from .framework import config

DIRECT_URL_NAME = env_var_prefix + 'DIRECT_URL'
SCRIPT_DIR_NAME = env_var_prefix + 'SCRIPT_DIR'

SKIP_JOB_LOAD_NAME = env_var_prefix + 'SKIP_JOB_LOAD'
SKIP_JOB_DELETE_NAME = env_var_prefix + 'SKIP_JOB_DELETE'


class ApiType(Enum):
    JENKINS = 0
    SCRIPT = 1
    MOCK = 2


_speedup = 1


def direct_url(api_type):
    if api_type != ApiType.SCRIPT:
        durl = os.environ.get(DIRECT_URL_NAME)
        return 'http://localhost:8080' if durl is None else durl.rstrip('/')
    else:
        return script_dir()


def public_url(api_type):
    if api_type != ApiType.SCRIPT:
        purl = os.environ.get('JENKINS_URL') or os.environ.get('HUDSON_URL')
        return "http://" + socket.getfqdn() + ':' + repr(8080) + '/' if purl is None else purl
    else:
        return script_dir()


def direct_cli_url(api_type):
    if api_type != ApiType.SCRIPT:
        purl = os.environ.get('JENKINS_URL')
        if purl:
            return direct_url(api_type) + '/jnlpJars/jenkins-cli.jar'
        purl = os.environ.get('HUDSON_URL')
        if purl:
            return direct_url(api_type) + '/jnlpJars/hudson-cli.jar'
        # If neither JENKINS nor HUDSON URL is set, assume jenkins for testing
        return direct_url(api_type) + '/jnlpJars/jenkins-cli.jar'
    else:
        return script_dir()


def public_cli_url(api_type):
    if api_type != ApiType.SCRIPT:
        purl = os.environ.get('JENKINS_URL')
        if purl:
            return purl.rstrip('/') + '/jnlpJars/jenkins-cli.jar'
        purl = os.environ.get('HUDSON_URL')
        if purl:
            return purl.rstrip('/') + '/jnlpJars/hudson-cli.jar'
        # If neither JENKINS nor HUDSON URL is set, assume jenkins for testing
        return "http://" + socket.getfqdn() + ':' + repr(8080) + '/jnlpJars/jenkins-cli.jar'
    else:
        return script_dir()


proxied_public_url = "http://myproxy.mydom/jenkins"


def proxied_public_cli_url(api_type):
    # Not required to be a real url
    if api_type != ApiType.SCRIPT:
        # If neither JENKINS nor HUDSON URL is set, assume jenkins for testing
        cli = '/jnlpJars/hudson-cli.jar' if os.environ.get('HUDSON_URL') else '/jnlpJars/jenkins-cli.jar'
        return proxied_public_url + cli
    else:
        return script_dir()


def script_dir():
    sdir = os.environ.get(SCRIPT_DIR_NAME)
    return config.job_script_dir if sdir is None else sdir.rstrip('/')


def select_speedup(speedup):
    """speedup is used by the mock api"""
    global _speedup
    _speedup = speedup


def speedup():
    return _speedup


def skip_job_delete():
    return os.environ.get(SKIP_JOB_DELETE_NAME) == 'true'


def skip_job_load():
    return os.environ.get(SKIP_JOB_LOAD_NAME) == 'true'


def skip_job_load_sh_export_str():
    return 'export ' + SKIP_JOB_LOAD_NAME + '=true'



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial, parallel
from .framework import api_select


def test_invoke_unchecked_dont_wait_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_slow_unchecked', max_fails=0, expect_invocations=1, expect_order=1, exec_time=100, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=2)

        with serial(api, timeout=50, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_slow_unchecked')
            ctrl1.invoke('j12')


def test_invoke_unchecked_dont_wait_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_slow_unchecked', max_fails=0, expect_invocations=1, expect_order=1, exec_time=100, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=2, exec_time=5)

        with parallel(api, timeout=50, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_slow_unchecked')
            ctrl1.invoke('j12')


def test_invoke_unchecked_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('j11_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=1, exec_time=5)
        api.job('j13_unchecked', max_fails=0, expect_invocations=1, expect_order=2, exec_time=30, invocation_delay=0, unknown_result=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12')
            ctrl1.invoke_unchecked('j13_unchecked')


def test_invoke_unchecked_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('j11_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=1, exec_time=5)
        api.job('j13_unchecked', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12')
            ctrl1.invoke_unchecked('j13_unchecked')


def test_invoke_unchecked_serial_fails(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('j11_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j13_fail_unchecked', max_fails=1, expect_invocations=1, expect_order=2)
        api.job('j14', max_fails=0, expect_invocations=1, expect_order=2, exec_time=5)
        api.job('j15_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12')
            ctrl1.invoke_unchecked('j13_fail_unchecked')
            ctrl1.invoke('j14')
            ctrl1.invoke_unchecked('j15_unchecked')


def test_invoke_unchecked_parallel_fails(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('j11_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j13_fail_unchecked', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('j14', max_fails=0, expect_invocations=1, expect_order=1, exec_time=5)
        api.job('j15_unchecked', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12')
            ctrl1.invoke_unchecked('j13_fail_unchecked')
            ctrl1.invoke('j14')
            ctrl1.invoke_unchecked('j15_unchecked')


def test_invoke_unchecked_mix_fails(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j31', max_fails=0, expect_invocations=1, expect_order=3)
        # Make sure result is available during first invocation of _check, only way to hit error handling code in unchecked job
        vfast = 0.00000000000000000000000000000000001
        api.job('j32_fail_unchecked', max_fails=1, expect_invocations=1, expect_order=3, exec_time=vfast, invocation_delay=0)
        api.job('j33_slow_unchecked', max_fails=0, expect_invocations=1, expect_order=None, exec_time=100, unknown_result=True)
        api.job('j34', max_fails=0, expect_invocations=1, expect_order=3, exec_time=5)
        api.job('j35_fail_unchecked', max_fails=1, expect_invocations=1, expect_order=3)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=4)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12')

            with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                    ctrl3a.invoke('j31')
                    ctrl3a.invoke_unchecked('j32_fail_unchecked')

                with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                    ctrl3b.invoke_unchecked('j33_slow_unchecked')
                    ctrl3b.invoke('j34')
                    ctrl3b.invoke_unchecked('j35_fail_unchecked')

            ctrl1.invoke('j13')


def test_invoke_unchecked_mix_no_fails(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('j31_unchecked', max_fails=0, expect_invocations=1, expect_order=1, exec_time=30, unknown_result=True)
        api.job('j32_unchecked', max_fails=0, expect_invocations=1, expect_order=1, exec_time=30, unknown_result=True)
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=2)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                    ctrl3a.invoke_unchecked('j31_unchecked')

                with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                    ctrl3b.invoke_unchecked('j32_unchecked')

            ctrl1.invoke('j11')



import sys
import os.path


error_msg = """
`{pkg}` 4.0+ supports Python {min_sup_version} and above.
When using Python 2.7, 3.4 - 3.5.x please install {pkg} 3.x

Python {py} detected.

Make sure you have pip >= 9.0 as well as setuptools >= 24.2 to avoid these kinds of issues:

 $ pip install pip setuptools --upgrade

Your choices:

- Upgrade to Python {min_sup_version}+.

- Install an older version of {pkg}:

 $ pip install '{pkg}<9.0'

It would be great if you can figure out how this version ended up being
installed, and try to check how to prevent that for future users.

Source: https://github.com/lhupfeldt/{pkg}
"""

min_sup_version = (3, 6, 0)
if sys.version_info < min_sup_version:
    raise ImportError(error_msg.format(
        py='.'.join([str(vv) for vv in sys.version_info[:3]]),
        pkg=os.path.basename(os.path.dirname(os.path.abspath(__file__))),
        min_sup_version='.'.join([str(vv) for vv in min_sup_version])))



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys

from .framework import api_select
from jenkinsflow.flow import serial
from jenkinsflow.unbuffered import UnBuffered


def test_unbuffered(api_type):
    sys.stdout = UnBuffered(sys.stdout)
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('unbuf', max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl:
            ctrl.invoke('unbuf')

    # TODO test output
    assert hasattr(sys.stdout, 'write') == True



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import serial, parallel, FailedChildJobException, FailedChildJobsException, FinalResultException, Propagation, BuildResult
from .framework import api_select


def test_propagation_warn_only_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_fail', max_fails=1, expect_invocations=1, expect_order=2, serial=True)
        api.job('j13', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FinalResultException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3, propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
                ctrl1.invoke('j11')
                ctrl1.invoke('j12_fail')
                ctrl1.invoke('j13')

        assert exinfo.value.result == BuildResult.UNSTABLE
        assert ctrl1.result == exinfo.value.result
        return 77


def test_propagation_warn_only_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1)

        with raises(FinalResultException) as exinfo:
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3, propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
                ctrl1.invoke('j1_fail')
                ctrl1.invoke('j2')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_serial_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=2)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=2)

        with raises(FinalResultException) as exinfo:
           with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
               ctrl1.invoke('j11')

               with ctrl1.parallel(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl2:
                   ctrl2.invoke('j21')
                   ctrl2.invoke('j22_fail')
                   ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_parallel_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=1, serial=True)
        api.job('j23', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FinalResultException) as exinfo:
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl2:
                    ctrl2.invoke('j21')
                    ctrl2.invoke('j22_fail')
                    ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_serial_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=2)
        api.job('j23', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FinalResultException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl2:
                    ctrl2.invoke('j21')
                    ctrl2.invoke('j22_fail')
                    ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_parallel_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=1)

        with raises(FinalResultException) as exinfo:
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3, raise_if_unsuccessful=True) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.parallel(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl2:
                    ctrl2.invoke('j21')
                    ctrl2.invoke('j22_fail')
                    ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_serial_serial_continue(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=3)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=4)

        with raises(FinalResultException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j21')
                    with ctrl2.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl3:
                        ctrl3.invoke('j22_fail')
                    ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_parallel_serial_continue(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=1)

        with raises(FinalResultException) as exinfo:
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j21')
                    with ctrl2.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl3:
                        ctrl3.invoke('j22_fail')
                    ctrl2.invoke('j23')

        assert exinfo.value.result == BuildResult.UNSTABLE
        return 77


def test_propagation_warn_only_nested_serial_serial_continue_fail(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=3)
        api.job('j23_fail', max_fails=1, expect_invocations=1, expect_order=4)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j21')
                    with ctrl2.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl3:
                        ctrl3.invoke('j22_fail')
                    ctrl2.invoke('j23_fail')


def test_propagation_warn_only_nested_parallel_serial_continue_fail(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j22_fail', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('j23_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with raises(FailedChildJobsException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('j11')

                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j21')
                    with ctrl2.serial(propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl3:
                        ctrl3.invoke('j22_fail')
                    ctrl2.invoke('j23_fail')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import re
from pytest import raises

from jenkinsflow.flow import parallel, serial, FailedChildJobException, FailedChildJobsException
from .framework import api_select
from .framework.utils import lines_in


def test_single_level_errors_parallel(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))
        api.job('quick_fail', max_fails=1, expect_invocations=1, expect_order=1)
        api.job('wait10', max_fails=0, expect_invocations=1, expect_order=1, exec_time=10)
        api.job('wait10_fail', max_fails=1, expect_invocations=1, expect_order=1, exec_time=10)
        api.job('wait5', max_fails=0, expect_invocations=1, expect_order=1, exec_time=5)
        api.job('wait5_fail', max_fails=1, expect_invocations=1, expect_order=1, exec_time=5)

        with raises(FailedChildJobsException) as exinfo:
            with parallel(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
                ctrl.invoke('quick', s1='', c1=False)
                ctrl.invoke('quick_fail')
                ctrl.invoke('wait10')
                ctrl.invoke('wait10_fail')
                ctrl.invoke('wait5')
                ctrl.invoke('wait5_fail')

        assert "quick_fail" in str(exinfo.value)
        assert "wait10_fail" in str(exinfo.value)
        assert "wait5_fail" in str(exinfo.value)

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            re.compile("^FAILURE: 'jenkinsflow_test__single_level_errors_parallel__quick_fail' - build: .*/jenkinsflow_test__single_level_errors_parallel__quick_fail.* after:"),
        )


def test_single_level_errors_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))
        api.job('quick_fail', max_fails=1, expect_invocations=1, expect_order=2, serial=True)
        api.job('wait5', max_fails=0, expect_invocations=0, expect_order=None, exec_time=5)

        with raises(FailedChildJobException):
            with serial(api, timeout=20, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
                ctrl.invoke('quick', s1='', c1=False)
                ctrl.invoke('quick_fail')
                ctrl.invoke('wait5')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import time

import pytest

from ..cfg import ApiType
from .hyperspeed import HyperSpeed


def test_hyperspeed_speedup():
    hs = HyperSpeed(1000)
    assert hs.speedup == 1000


@pytest.mark.apis(ApiType.MOCK)
def test_hyperspeed_mocked_time():
    hs = HyperSpeed(1000)
    time.sleep(0.001)
    assert hs.time() > time.time()


@pytest.mark.apis(ApiType.MOCK)
def test_hyperspeed_mocked_sleep():
    hs = HyperSpeed(1000)
    before = time.time()
    hs.sleep(1)
    after = time.time()
    assert after - before < 1



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, os
from os.path import join as jp
import subprocess

import pytest
from pytest import raises

from jenkinsflow.flow import serial
from jenkinsflow.utils.set_build_description import set_build_description
from jenkinsflow.cli.cli import cli

from demo_security import username, password, set_build_description_must_authenticate

from .framework import api_select
from . import cfg as test_cfg
from .cfg import ApiType


_here = os.path.dirname(os.path.abspath(__file__))


def _clear_description(api, job):
    if api.api_type == ApiType.SCRIPT:
        # TODO: There is no build number concept for script api, so we need to ensure a clean start
        description_file = jp(job.workspace, 'description.txt')
        if os.path.exists(description_file):
            os.remove(description_file)


def _get_description(api, job, build_number):
    # Read back description and verify
    if api.api_type == ApiType.JENKINS:
        build_url = "/job/" + job.name + '/' + str(build_number)
        dct = api.get_json(build_url, tree="description")
        return dct['description']

    if api.api_type == ApiType.SCRIPT:
        with open(jp(job.workspace, 'description.txt')) as ff:
            return ff.read()

    if api.api_type == ApiType.MOCK:
        raise Exception(repr(ApiType.MOCK) + " 'description' not supported by test framework")


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_flow_set(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('job-1', max_fails=0, expect_invocations=1, expect_order=1, params=_params)
        api.job('job-2', max_fails=0, expect_invocations=1, expect_order=2, params=_params, serial=True)
        api.job('job-3', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-4', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-5', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-6', max_fails=0, expect_invocations=1, expect_order=4, params=_params)
        api.job('job-7', max_fails=0, expect_invocations=1, expect_order=5, params=_params, serial=True)

        if api.api_type == ApiType.SCRIPT:
            for job_num in range(1, 7):
                job = api.get_job(api.job_name_prefix + 'job-' + str(job_num))
                _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="AAA") as ctrl1:
            ctrl1.invoke('job-1', password='a', s1='b')
            ctrl1.invoke('job-2', password='a', s1='b')

            with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                    ctrl3a.invoke('job-3', password='a', s1='b')
                    ctrl3a.invoke('job-6', password='a', s1='b')

                with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                    ctrl3b.invoke('job-4', password='a', s1='b')
                    ctrl3b.invoke('job-5', password='a', s1='b')

            ctrl1.invoke('job-7', password='a', s1='b')

        for job_num in range(1, 7):
            job = api.get_job(api.job_name_prefix + 'job-' + str(job_num))
            _, _, build_num = job.job_status()
            assert _get_description(api, job, build_num) == 'AAA'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_api(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        # Need to read the build number
        if api.api_type == ApiType.SCRIPT:
            # TODO: This can't be called here for Jenkins API. Why?
            job = api.get_job(api.job_name_prefix + job_name)
            _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="AAA") as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        if api.api_type != ApiType.SCRIPT:
            job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        api.set_build_description('BBB1', job_name=job.name, build_number=build_num)
        api.set_build_description('BBB2', False, job_name=job.name, build_number=build_num)
        assert _get_description(api, job, build_num) == 'AAA\nBBB1\nBBB2'

        api.set_build_description('BBB3', replace=True, job_name=job.name, build_number=build_num)
        api.set_build_description('BBB4', replace=False, separator='#', job_name=job.name, build_number=build_num)
        api.set_build_description('BBB5', separator='!!', job_name=job.name, build_number=build_num)
        assert _get_description(api, job, build_num) == 'BBB3#BBB4!!BBB5'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_utils(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        # Need to read the build number
        if api.api_type == ApiType.SCRIPT:
            # TODO: This can't be called here for Jenkins API. Why?
            job = api.get_job(api.job_name_prefix + job_name)
            _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="AAA") as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        if api.api_type != ApiType.SCRIPT:
            job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()
        direct_url = test_cfg.direct_url(api_type)

        if set_build_description_must_authenticate:
            set_build_description('BBB1', job_name=job.name, build_number=build_num, direct_url=direct_url, username=username, password=password)
            set_build_description('BBB2', replace=False, job_name=job.name, build_number=build_num, direct_url=direct_url, username=username, password=password)
        else:
            set_build_description('BBB1', job_name=job.name, build_number=build_num, direct_url=direct_url)
            set_build_description('BBB2', replace=False, job_name=job.name, build_number=build_num, direct_url=direct_url)

        assert _get_description(api, job, build_num) == 'AAA\nBBB1\nBBB2'

        set_build_description('BBB3', replace=True, job_name=job.name, build_number=build_num, direct_url=direct_url, username=username, password=password)
        set_build_description(
            'BBB4', replace=False, separator='#', job_name=job.name, build_number=build_num, direct_url=direct_url, username=username, password=password)
        set_build_description('BBB5', separator='!!', job_name=job.name, build_number=build_num, direct_url=direct_url, username=username, password=password)
        assert _get_description(api, job, build_num) == 'BBB3#BBB4!!BBB5'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_utils_from_inside_job_1(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1,
                set_build_descriptions=(('BBB1', False, '\\n'), ('BBB2', False, '\\n')))

        # Need to read the build number
        if api.api_type == ApiType.SCRIPT:
            # TODO: This can't be called here for Jenkins API. Why?
            job = api.get_job(api.job_name_prefix + job_name)
            _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="AAA") as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        if api.api_type != ApiType.SCRIPT:
            job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        assert _get_description(api, job, build_num) == 'AAA\nBBB1\nBBB2'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_utils_from_inside_job_2(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1,
                set_build_descriptions=(('BBB3', True, '\\n'), ('BBB4', False, '#'), ('BBB5', False, '!!')))

        # Need to read the build number
        if api.api_type == ApiType.SCRIPT:
            # TODO: This can't be called here for Jenkins API. Why?
            job = api.get_job(api.job_name_prefix + job_name)
            _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="AAA") as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        if api.api_type != ApiType.SCRIPT:
            job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        assert _get_description(api, job, build_num) == 'BBB3#BBB4!!BBB5'


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_set_build_description_unknown_job(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        job_name = 'job-1'

        with raises(Exception) as exinfo:
            api.set_build_description('Oops', job_name=job_name, build_number=17)

        assert "Build not found " in str(exinfo.value)


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_cli(api_type, cli_runner):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        # Need to read the build number
        job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()
        base_url = test_cfg.direct_url(api_type) + '/'

        _clear_description(api, job)

        cli_args = [
            'set_build_description',
            '--job-name', job.name,
            '--build-number', repr(build_num),
            '--description', 'BBB1',
            '--direct-url', base_url,
            '--separator', '\n',
            '--username', username,
            '--password', password]
        print("cli args:", cli_args)

        result = cli_runner.invoke(cli, cli_args)
        print(result.output)
        assert not result.exception
        assert _get_description(api, job, build_num) == 'BBB1'

        cli_args = [
            'set_build_description',
            '--job-name', job.name,
            '--build-number', repr(build_num),
            '--description', 'BBB2',
            '--direct-url', base_url,
            '--replace',
            '--username', username,
            '--password', password]
        print("cli args:", cli_args)

        result = cli_runner.invoke(cli, cli_args)
        print(result.output)
        assert not result.exception
        assert _get_description(api, job, build_num) == 'BBB2'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_cli_env_url(api_type, env_base_url, cli_runner):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'j1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        # Need to read the build number
        job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        _clear_description(api, job)

        cli_args = [
            'set_build_description',
            '--job-name', job.name,
            '--build-number', repr(build_num),
            '--description', 'BBB1',
            '--separator', '\n',
            '--username', username,
            '--password', password]
        print("cli args:", cli_args)

        result = cli_runner.invoke(cli, cli_args)
        print(result.output)
        assert not result.exception
        assert _get_description(api, job, build_num) == 'BBB1'


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_description_cli_no_env_url(api_type, env_no_base_url, cli_runner):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'j1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        # Need to read the build number
        job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        cli_args = [
            'set_build_description',
            '--job-name', job.name,
            '--build-number', repr(build_num),
            '--description', 'BBB1']
        print("cli args:", cli_args)

        result = cli_runner.invoke(cli, cli_args)
        print(result.output)
        assert result.exception
        assert "Could not get env variable JENKINS_URL or HUDSON_URL" in str(result.exception)
        assert "You must specify 'direct-url'" in result.output


def test_set_build_description_call_script_help(capfd):
    # Invoke this in a subprocess to ensure that calling the script works
    # This will not give coverage as it not not traced through the subprocess call
    rc = subprocess.call([sys.executable, jp(_here, '../cli/cli.py'), 'set_build_description', '--help'])
    assert rc == 0

    sout, _ = capfd.readouterr()
    assert '--job-name' in sout
    assert '--build-number' in sout
    assert '--description' in sout
    assert '--direct-url' in sout
    assert '--replace' in sout
    assert '--separator' in sout
    assert '--username' in sout
    assert '--password' in sout



import sys


class _NotFoundException(Exception):
    pass


def lines_in(text, mfunc, *expected_lines):
    """Test that `*expected_lines` occur in order in the lines of `text`.

    Arguments:
        text (str): The text to find *expected_lines in
        mfunc (callable(str) -> str): Called from every line in text and every str (non re) in expected_lines.
            Can be used to modify the each str before comparison.
        *expected_lines (str, RegexObject (hasattr `match`) or sequence): For each `expected_line` in expected_lines:
            If an expected_line is a tuple or a list, any item in the sequence is handled as an individual
            expected_line, which may be matched in any order, but not out of order with the surrounding expected_lines,
            as described below:

            If `expected_line` has a match method it is called and must return True for a line in `text`.
            Otherwise, if the `expected_line` starts with '^', a line in `text` must start with `expected_line[1:]`
            Otherwise `expected line` must simply occur in a line in `text`

    Return (bool): True if lines found in order, else False
    """

    text_lines = text.split('\n')
    max_line_index = len(text_lines)
    matched_lines = []

    def _check_match(expected, mline, line):
        expected = mfunc(expected) if mfunc and not hasattr(expected, 'match') else expected

        if hasattr(expected, 'match'):
            if expected.match(mline):
                matched_lines.append((line, mline))
                return True
            return False

        if expected.startswith('^'):
            if mline.startswith(expected[1:]):
                matched_lines.append((line, mline))
                return True
            return False

        if expected in mline:
            matched_lines.append((line, mline))
            return True

        return False

    def _report_failure(expected):
        num_matched = len(matched_lines)
        if num_matched:
            lines = []
            for line, mline in matched_lines:
                lines.append((line + ' (' + mline + ')') if line != mline else line)
            matched = "Matched {num} lines{mfunc_info}:\n\n{lines}".format(
                num=num_matched, mfunc_info=" ('mfunc' modified line in '()')" if mfunc else "", lines='\n'.join(lines))
        else:
            matched = 'No lines matched.' + " (Empty text)" if not text else ''

        raw_txt_msg = " (unmodified text, 'mfunc' not applied)" if mfunc else ''
            
        if hasattr(expected, 'match'):
            print('', matched, "The regex:", expected.pattern,
                  "--- NOT MATCHED or OUT OF ORDER in" + raw_txt_msg + " ---", text, file=sys.stderr, sep='\n\n')
            raise _NotFoundException()

        if expected.startswith('^'):
            expected = expected[1:]
            expinfo = expected + (" (" + mfunc(expected) + ")" if mfunc else "")
            print('', matched, "The text:", expinfo,
                  "--- NOT FOUND, OUT OF ORDER or NOT AT START OF LINE in" + raw_txt_msg + " ---", text, file=sys.stderr, sep='\n\n')
            raise _NotFoundException()

        expinfo = expected + (" (" + mfunc(expected) + ")" if mfunc else "")
        print('', matched, "The text:", expinfo,
              "--- NOT FOUND OR OUT OF ORDER IN" + raw_txt_msg + " ---", text, file=sys.stderr, sep='\n\n')
        raise _NotFoundException()

    def _ordered_lines_in(text_line_index, mfunc, expected_lines):
        max_expected_index = len(expected_lines)
        expected_index = 0

        while True:
            if expected_index == max_expected_index:
                return text_line_index
            expected = expected_lines[expected_index]

            if text_line_index == max_line_index:
                _report_failure(expected)

            if isinstance(expected, (tuple, list)):
                text_line_index = _unordered_lines_in(text_line_index, mfunc, expected)
                expected_index += 1
                continue

            line = text_lines[text_line_index]
            mline = mfunc(line) if mfunc else line
            if _check_match(expected, mline, line):
                expected_index += 1

            text_line_index += 1

    def _unordered_lines_in(text_line_index, mfunc, expected_lines):
        expected_lines = list(expected_lines)
        expected_index = 0

        while True:
            line = text_lines[text_line_index]
            mline = mfunc(line) if mfunc else line

            for expected_index, expected in enumerate(expected_lines):
                if isinstance(expected, (tuple, list)):
                    text_line_index = _ordered_lines_in(text_line_index, mfunc, expected)
                    del expected_lines[expected_index]
                    break

                if _check_match(expected, mline, line):
                    del expected_lines[expected_index]
                    break

            text_line_index += 1
            if not expected_lines:
                return text_line_index

            if text_line_index == max_line_index:
                _report_failure(expected)

    try:
        _ordered_lines_in(0, mfunc, expected_lines)
        return True
    except _NotFoundException:
        return False



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, os, imp
from os.path import join as jp
here = os.path.abspath(os.path.dirname(__file__))

extra_sys_path = [os.path.normpath(path) for path in [here, jp(here, '../..'), jp(here, '../demo'), jp(here, '../demo/jobs')]]
sys.path = extra_sys_path + sys.path

import pytest
from pytest import raises

from jenkinsflow.flow import parallel, JobControlFailException
from .framework import api_select
from .cfg import ApiType

import basic, calculated_flow, prefix, hide_password, errors


def load_demo_jobs(demo, api_type):
    print("\nLoad jobs for demo:", demo.__name__)
    job_load_module_name = demo.__name__ + '_jobs'
    job_load = imp.load_source(job_load_module_name, jp(here, '../demo/jobs', job_load_module_name + '.py'))
    api = job_load.create_jobs(api_type)
    return api


def _test_demo(demo, api_type):
    load_demo_jobs(demo, api_type)

    with api_select.api(__file__, api_type, fixed_prefix="jenkinsflow_demo__") as api:
        api.job(demo.__name__ + "__0flow", max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke(demo.__name__ + "__0flow")


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_basic(api_type):
    _test_demo(basic, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_calculated_flow(api_type):
    _test_demo(calculated_flow, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_prefix(api_type):
    _test_demo(prefix, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_hide_password(api_type):
    _test_demo(hide_password, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_with_errors(api_type):
    demo = errors
    load_demo_jobs(demo, api_type)

    with api_select.api(__file__, api_type, fixed_prefix="jenkinsflow_demo__") as api:
        api.job(demo.__name__ + "__0flow", max_fails=1, expect_invocations=1, expect_order=1)

        with raises(JobControlFailException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke(demo.__name__ + "__0flow")



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, time, re

import pytest
from pytest import raises, xfail

from jenkinsflow.flow import serial, FailedChildJobException, FinalResultException, BuildResult

from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in, kill_current_msg
from .framework.killer import kill

here = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO
def test_kill_all_unchecked(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, invocation_delay=0, unknown_result=False, kill=True)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1, invocation_delay=0, unknown_result=False, kill=True)
        api.job('j3', max_fails=0, expect_invocations=3, expect_order=None, exec_time=50, invocation_delay=0, unknown_result=False, kill=True, allow_running=True)
        api.job('j4', max_fails=0, expect_invocations=1, expect_order=None, invocation_delay=0, unknown_result=False, kill=True)
        api.job('j5', max_fails=1, expect_invocations=1, expect_order=2, invocation_delay=0, unknown_result=False, kill=True)

        def flow(api, kill_all):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, kill_all=kill_all) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    ctrl2.invoke_unchecked('j1')
                    ctrl2.invoke_unchecked('j2')
                with ctrl1.parallel() as ctrl3:
                    ctrl3.invoke_unchecked('j3')
                    ctrl3.invoke_unchecked('j3') # Queue
                    ctrl3.invoke_unchecked('j3') # Queue
                    ctrl3.invoke_unchecked('j4')
                    ctrl3.invoke_unchecked('j5')

        # Invoke the flow
        flow(api, False)

        # Make sure job has actually started before entering new flow
        api.sleep(5)

        if capsys:
            sout, _ = capsys.readouterr()
            assert lines_in(api_type, sout, "unchecked job: 'jenkinsflow_test__kill_all_unchecked__j1' UNKNOWN - RUNNING")
            assert lines_in(api_type, sout, "unchecked job: 'jenkinsflow_test__kill_all_unchecked__j3' Invocation-1 UNKNOWN - RUNNING")

        # Kill the flow
        flow(api, True)

        if not capsys:
            # Not called by pytest, but from flow hudson job
            return

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Killing all running builds for: 'jenkinsflow_test__kill_all_unchecked__j1'",
            "job: 'jenkinsflow_test__kill_all_unchecked__j1' stopped running",
            "job: 'jenkinsflow_test__kill_all_unchecked__j1' Status IDLE - build: ",
            "^ABORTED: 'jenkinsflow_test__kill_all_unchecked__j1'",
            "^--- Final status ---",
            "^serial flow: [",
            "^   parallel flow: (",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j1' ABORTED - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j2' SUCCESS - IDLE",
            "^   )",

            "^   parallel flow: (",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j3' Invocation-1 ABORTED - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j3' Invocation-2 ABORTED - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j3' Invocation-3 ABORTED - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j4' SUCCESS - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked__j5' FAILURE - IDLE",
            "^   )",
            "^]",
        )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_kill_mini(api_type, capsys):
    """Cut down kill_current for debugging"""
    with api_select.api(__file__, api_type, login=True) as api:
        is_hudson = os.environ.get('HUDSON_URL')
        if is_hudson:  # TODO investigate why this test fails in Hudson
            xfail("Doesn't pass in Hudson")
            return

        api.flow_job()
        num_j6_invocations = 2
        api.job('j6', max_fails=0, expect_invocations=num_j6_invocations, expect_order=None, exec_time=50, kill=True,
                num_builds_to_keep=num_j6_invocations*2 + 1, params=(('a', 0, 'integer'),))

        kill(api, 35, 1)

        with raises(FinalResultException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.05) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    for ii in range(0, num_j6_invocations):
                         # Queue a lot of jobs
                        ctrl2.invoke('j6', a=ii)

        assert exinfo.value.result == BuildResult.FAILURE  # TODO? Note that ABORTED on child job propagates as FAILURE to the flow

        if not capsys:
            return

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Got SIGTERM: Killing all builds belonging to current flow",
            "^--- Final status ---",
            "^serial flow: [",
            "^   parallel flow: (",
            re.compile("      job: 'jenkinsflow_test__kill_mini__j6' Invocation-[0-9]+ ABORTED - IDLE"),
            re.compile("      job: 'jenkinsflow_test__kill_mini__j6' Invocation-[0-9]+ DEQUEUED - IDLE"),
            "^   )",
            "^",
            "^]"
        )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)  # TODO
def test_kill_current(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        is_hudson = os.environ.get('HUDSON_URL')
        if is_hudson:  # TODO investigate why this test fails in Hudson
            xfail("Doesn't pass in Hudson")
            return

        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, kill=True)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j3', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, kill=True)
        api.job('j4', max_fails=1, expect_invocations=1, expect_order=2)
        api.job('j5', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, kill=True)

        num_j6_invocations = 20
        api.job('j6', max_fails=0, expect_invocations=num_j6_invocations, expect_order=None, exec_time=50, kill=True,
                num_builds_to_keep=num_j6_invocations*2 + 1, params=(('a', 0, 'integer'),))
        api.job('j7', max_fails=0, expect_invocations=0, expect_order=None, exec_time=50)

        # Set a long sleep here, when heaviliy loaded it can take time for the flow to get started
        kill(api, 35, 1)

        # TODO: shouldn't we expect a FailedChildJobsException here?
        with raises(FinalResultException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.05) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    ctrl2.invoke('j1')
                    with ctrl2.serial() as ctrl3:
                        ctrl3.invoke('j2')
                        ctrl3.invoke('j3')
                    ctrl2.invoke('j4')
                    ctrl2.invoke_unchecked('j5')
                    for ii in range(0, num_j6_invocations):
                         # Queue a lot of jobs
                        ctrl2.invoke('j6', a=ii)
                with ctrl1.parallel() as ctrl4:
                    ctrl4.invoke('j7')

        assert exinfo.value.result == BuildResult.FAILURE  # TODO? Note that ABORTED on child job propagates as FAILURE to the flow

        if not capsys:
            return

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Got SIGTERM: Killing all builds belonging to current flow",
            kill_current_msg(api, 'jenkinsflow_test__kill_current__j1', 1),
            kill_current_msg(api, 'jenkinsflow_test__kill_current__j3', 1),
            kill_current_msg(api, 'jenkinsflow_test__kill_current__j5', 1),
            "^--- Final status ---",
            "^serial flow: [",
            "^   parallel flow: (",
            "^      job: 'jenkinsflow_test__kill_current__j1' ABORTED - IDLE",
            "^      serial flow: [",
            "^         job: 'jenkinsflow_test__kill_current__j2' SUCCESS - IDLE",
            "^         job: 'jenkinsflow_test__kill_current__j3' ABORTED - IDLE",
            "^      ]",
            "^      job: 'jenkinsflow_test__kill_current__j4' FAILURE - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_current__j5' ABORTED - IDLE",
            re.compile("      job: 'jenkinsflow_test__kill_current__j6' Invocation-[0-9]+ ABORTED - IDLE"),
            re.compile("      job: 'jenkinsflow_test__kill_current__j6' Invocation-[0-9]+ DEQUEUED - IDLE"),
            "^   )",
            "^",
            "^   parallel flow: (",
            "^      job: 'jenkinsflow_test__kill_current__j7' UNKNOWN - IDLE",
            "^   )",
            "^",
            "^]"
        )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)  # TODO
def test_kill_all_unchecked_no_job(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, unknown_result=False, kill=True)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1, unknown_result=False, kill=True)
        #api.job('j3', 0, 0, None, non_existing=True)
        api.job('j4', max_fails=0, expect_invocations=1, expect_order=1, unknown_result=False, kill=True)
        #api.job('j5', 0, 0, None, non_existing=True)

        def flow(api, kill_all, allow_missing_jobs):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, allow_missing_jobs=allow_missing_jobs, kill_all=kill_all) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    ctrl2.invoke_unchecked('j1')
                    ctrl2.invoke_unchecked('j2')
                with ctrl1.parallel() as ctrl3:
                    ctrl3.invoke('j3')
                    ctrl3.invoke('j4')
                    ctrl3.invoke('j5')

        # Invoke the flow
        with raises(FailedChildJobException):
            flow(api, False, True)

        # Make sure job has actually started before entering new flow
        api.sleep(5)

        if capsys:
            sout, _ = capsys.readouterr()
            assert lines_in(api_type, sout, "unchecked job: 'jenkinsflow_test__kill_all_unchecked_no_job__j1' UNKNOWN - RUNNING")
            assert lines_in(api_type, sout, "job: 'jenkinsflow_test__kill_all_unchecked_no_job__j3' - MISSING JOB")

        # Kill the flow
        flow(api, True, False)

        if not capsys:
            return

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Killing all running builds for: 'jenkinsflow_test__kill_all_unchecked_no_job__j1'",
            "job: 'jenkinsflow_test__kill_all_unchecked_no_job__j1' stopped running",
            "job: 'jenkinsflow_test__kill_all_unchecked_no_job__j1' Status IDLE - build: ",
            "^ABORTED: 'jenkinsflow_test__kill_all_unchecked_no_job__j1'",
            "^--- Final status ---",
            "^serial flow: [",
            "^   parallel flow: (",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked_no_job__j1' ABORTED - IDLE",
            "^      unchecked job: 'jenkinsflow_test__kill_all_unchecked_no_job__j2' SUCCESS - IDLE",
            "^   )",

            "^   parallel flow: (",
            "^      job: 'jenkinsflow_test__kill_all_unchecked_no_job__j3' - MISSING JOB",
            "^      job: 'jenkinsflow_test__kill_all_unchecked_no_job__j4' SUCCESS - IDLE",
            "^      job: 'jenkinsflow_test__kill_all_unchecked_no_job__j5' - MISSING JOB",
            "^   )",
            "^]",
        )



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

# Abstract base classes representing the API's used from jenkins_apixk

# These classes serve as base classes for the Mock and Wrappper test API's

import abc


class AbstractApiJob(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def invoke(self, securitytoken, build_params, cause, description):
        raise Exception("AbstractNotImplemented")

    # The following should be declare abstract, but since they are 'implemented' by proxy we can't do that (conveniently)
    # def is_running(self):
    # def is_queued(self):
    # def get_last_build_or_none(self):
    # def get_build_triggerurl(self):
    # def update_config(self, config_xml):
    # def poll(self):


class AbstractApiJenkins(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def create_job(self, job_name, config_xml):
        raise Exception("AbstractNotImplemented")

    @abc.abstractmethod
    def delete_job(self, job_name):
        raise Exception("AbstractNotImplemented")

    @abc.abstractmethod
    def get_job(self, name):
        raise Exception("AbstractNotImplemented")

    @abc.abstractmethod
    def poll(self):
        raise Exception("AbstractNotImplemented")



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, os


def env_base_url():
    """Determines base_url when run from within a Jenkins/Hudson job

    Returns:
        base_url (str): If JENKINS_URL or HUDSON_URL is set, i.e. run from within job.

    Raises:
        Exception if neither JENKINS_URL nor HUDSON_URL is set, i.e. NOT run from within job.
    """

    base_url = os.environ.get('JENKINS_URL') or os.environ.get('HUDSON_URL')
    if base_url:
        return base_url.rstrip('/')

    if base_url is None:
        raise Exception("Could not get env variable JENKINS_URL or HUDSON_URL. "
                        "You must set 'Jenkins Location' in Jenkins setup for JENKINS_URL to be exported. "
                        "You must set 'Hudson URL' in Hudson setup for HUDSON_URL to be exported.")


def base_url_and_api(direct_url):
    try:
        base_url = direct_url.rstrip('/') if direct_url else env_base_url()
    except Exception:
        print("*** ERROR: You must specify 'direct-url' if not running from Jenkins job", file=sys.stderr)
        raise

    if base_url.startswith('http:') or base_url.startswith('https:'):
        # Using jenkins_api
        from .. import jenkins_api as api
    else:
        # Using script_api
        from .. import script_api as api

    return base_url, api



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys, os, shutil, importlib, datetime, tempfile, psutil, setproctitle, signal, errno
from os.path import join as jp
import multiprocessing

from .api_base import BuildResult, Progress, UnknownJobException, ApiInvocationMixin
from .speed import Speed


here = os.path.abspath(os.path.dirname(__file__))


def _mkdir(path):
    try:
        os.mkdir(path)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def _pgrep(proc_name):
    """Returns True if a process with name 'proc_name' is running, else False"""
    try:
        for proc in psutil.process_iter():
            if proc_name == proc.name():
                return True
    except psutil.NoSuchProcess:
        return False
    return False


class LoggingProcess(multiprocessing.Process):
    proc_name_prefix = "jenkinsflow_script_api_"

    def __init__(self, group=None, target=None, output_file_name=None, workspace=None, name=None, args=(), env=None):
        self.user_target = target
        super().__init__(group=group, target=self.run_job_wrapper, name=name, args=args)
        self.output_file_name = output_file_name
        self.workspace = workspace
        self.env = env
        self._build_res_unstable = False

    def run_job_wrapper(self, *args):
        setproctitle.setproctitle(self.proc_name_prefix + self.name)

        # Set signalhandler for changing job result
        def set_result(_sig, _frame):
            print("\nGot SIGUSR1: Changing result to 'unstable'")
            self._build_res_unstable = True
        signal.signal(signal.SIGUSR1, set_result)

        os.chdir(self.workspace)
        os.environ.update(self.env)
        os.environ['EXECUTOR_NUMBER'] = repr(self.pid)

        try:
            rc = self.user_target(*args)
        except Exception as ex:  # pylint: disable=broad-except
            print("jenkinsflow.script_api: Caught exception from job script:", ex)
            rc = 1

        if self._build_res_unstable:
            sys.exit(2)
        sys.exit(rc)

    def run(self):
        sys.stdout = sys.stderr = open(self.output_file_name, 'w', buffering=1)
        super().run()


class Jenkins(Speed):
    """Optimized minimal set of methods needed for jenkinsflow to directly execute python code instead of invoking Jenkins jobs.

    THIS DOES NOT SUPPORT CONCURRENT INVOCATIONS OF FLOW

    There is no concept of job queues or executors, so if your flow depends on these for correctness, you wil experience different behaviour
    when using this api instead of the real Jenkins.

    Args:
        direct_uri (str): Path to dir with 'job' method python modules. Modules named <job-name>.py will be imported from this directory.
            If no module exists for a specific jobname, the module called 'default.py' will be imported.
            The modules must contain at method called 'run_job' with the following signature:

                run_job(job_name, job_prefix_filter, username, password, securitytoken, cause, build_params)

                A return value of 0 is 'SUCCESS'
                A return value of 1 or any exception raised is 'FAILURE'
                Other return values means 'UNSTABLE'

        job_prefix_filter (str): Passed to 'run_job'. ``jenkinsflow`` puts no meaning into this parameter.
        username (str): Passed to 'run_job'. ``jenkinsflow`` puts no meaning into this parameter.
        password (str): Passed to 'run_job'. ``jenkinsflow`` puts no meaning into this parameter.
        invocation_class (class): Defaults to `Invocation`.
        log_dir (str): Directory in which to store logs. Defaults to subdirectory 'jenkinsflow' under the system defined tmp dir.
        **kwargs: Ignored for compatibility with the other jenkins apis
    """

    def __init__(self, direct_uri, job_prefix_filter=None, username=None, password=None, invocation_class=None, log_dir=None, **kwargs):
        self.job_prefix_filter = job_prefix_filter
        self.username = username
        self.password = password
        self.public_uri = direct_uri
        self.log_dir = log_dir or os.path.join(tempfile.gettempdir(), 'jenkinsflow')
        self.invocation_class = invocation_class or Invocation
        self.jobs = {}

    def poll(self):
        pass

    def quick_poll(self):
        pass

    def queue_poll(self):
        pass

    def _script_file(self, job_name):
        return jp(self.public_uri, job_name + '.py')

    def _workspace(self, job_name):
        return jp(self.public_uri, job_name)

    def get_job(self, name):
        job = self.jobs.get(name)
        if not job:
            script_file = script_file1 = self._script_file(name)
            if not os.path.exists(script_file):
                script_file = self._script_file('default')
                if not os.path.exists(script_file):
                    raise UnknownJobException(script_file1 + ' or ' + script_file)

            script_dir = os.path.dirname(script_file)
            if script_dir not in sys.path:
                sys.path.append(script_dir)

            try:
                user_module = importlib.import_module(os.path.basename(script_file).replace('.py', ''), package=None)
            except (ImportError, SyntaxError) as ex:
                raise UnknownJobException(repr(script_file) + ' ' + repr(ex))

            try:
                func = user_module.run_job
            except AttributeError as ex:
                raise UnknownJobException(script_file + repr(ex))
            job = self.jobs[name] = ApiJob(jenkins=self, name=name, script_file=script_file, workspace=self._workspace(name), func=func)
        return job

    def create_job(self, job_name, config_xml):
        script_file = self._script_file(job_name)
        _mkdir(os.path.dirname(script_file))
        with open(script_file, 'w') as ff:
            ff.write(config_xml)

    def delete_job(self, job_name):
        script_file = self._script_file(job_name)
        try:
            os.unlink(script_file)
        except OSError as ex:
            if not os.path.exists(script_file):
                raise UnknownJobException(script_file + repr(ex))
            raise

        try:
            shutil.rmtree(self._workspace(job_name))
        except OSError as ex:
            if os.path.exists(script_file):
                raise

    def set_build_description(self, description, replace=False, separator='\n', job_name=None, build_number=None):
        """Utility to set/append build description. :py:obj:`description` will be written to a file in the workspace.

        Args:
            description (str): The description to set on the build.
            append (bool): If True append to existing description, if any.
            separator (str): A separator to insert between any existing description and the new :py:obj:`description` if :py:obj:`append` is True.
            job_name (str):
            build_number (int):
        """

        if job_name is None:
            job_name = os.environ['JOB_NAME']

        if build_number is None:
            build_number = int(os.environ['BUILD_NUMBER'])

        workspace = self._workspace(job_name)
        mode = 'w' if replace else 'a'
        fname = jp(workspace, 'description.txt')
        if not replace and os.path.exists(fname) and os.stat(fname).st_size:
            description = separator + description
        with open(fname, mode) as ff:
            try:
                ff.write(description)
            except UnicodeEncodeError:
                ff.write(description.encode('utf-8'))


class ApiJob():
    def __init__(self, jenkins, name, script_file, workspace, func):
        self.jenkins = jenkins
        self.name = name

        self.build = None
        self.public_uri = script_file
        self.workspace = workspace
        self.func = func
        self.log_file = jp(self.jenkins.log_dir, self.name + '.log')
        self.build_num = None
        self._invocations = []
        self.queued_why = None
        self.old_build_number = None

    def invoke(self, securitytoken, build_params, cause, description):
        _mkdir(self.jenkins.log_dir)
        _mkdir(self.workspace)
        build_number = (self.build_num or 0) + 1
        self.build_num = build_number

        fixed_args = [self.name, self.jenkins.job_prefix_filter, self.jenkins.username, self.jenkins.password, securitytoken, cause]
        fixed_args.append(build_params if build_params else {})

        # Export some of the same variables that Jenkins does
        extra_env = dict(
            BUILD_NUMBER=repr(build_number),
            BUILD_ID=datetime.datetime.isoformat(datetime.datetime.utcnow()),
            BUILD_DISPLAY_NAME='#' + repr(build_number),
            JOB_NAME=self.name,
            BUILD_TAG='jenkinsflow-' + self.name + '-' + repr(build_number),
            NODE_NAME='master',
            NODE_LABELS='',
            WORKSPACE=self.workspace,
            JENKINS_HOME=self.jenkins.public_uri,
            JENKINS_URL=self.jenkins.public_uri,
            HUDSON_URL=self.jenkins.public_uri,
            BUILD_URL=jp(self.public_uri, repr(build_number)),
            JOB_URL=self.public_uri,
        )

        proc = LoggingProcess(target=self.func, output_file_name=self.log_file, workspace=self.workspace, name=self.name, args=fixed_args, env=extra_env)
        self.build = self.jenkins.invocation_class(self, proc, build_number)
        if description:
            self.jenkins.set_build_description(description, replace=True, separator='', job_name=self.name, build_number=build_number)
        self._invocations.append(self.build)
        return self.build

    def poll(self):
        pass

    def job_status(self):
        """Result, progress and latest buildnumber info for the JOB NOT the invocation

        Return (result, progress_info, latest_build_number) (str, str, int or None):
            Note: Always returns result == BuildResult.UNKNOWN and latest_build_number == 0
        """

        progress = Progress.RUNNING if _pgrep(LoggingProcess.proc_name_prefix + self.name) else Progress.IDLE
        result = BuildResult.UNKNOWN
        return (result, progress, 0)

    def stop_all(self):
        # TODO stop ALL
        if self.build:
            self.build.proc.terminate()

    def update_config(self, config_xml):
        _mkdir(os.path.dirname(self.public_uri))
        with open(self.public_uri, 'w') as ff:
            ff.write(config_xml)

    def __repr__(self):
        return str(self.name)


class Invocation(ApiInvocationMixin):
    def __init__(self, job, proc, build_number):
        self.job = job
        self.proc = proc
        self.build_number = build_number
        self.queued_why = None

        self.proc.start()

    def status(self):
        if self.proc.is_alive():
            return (BuildResult.UNKNOWN, Progress.RUNNING)
        rc = self.proc.exitcode
        if rc == 0:
            return (BuildResult.SUCCESS, Progress.IDLE)
        if rc == 1:
            return (BuildResult.FAILURE, Progress.IDLE)
        return (BuildResult.UNSTABLE, Progress.IDLE)

    def stop(self, dequeue):  # pylint: disable=unused-argument
        self.proc.terminate()

    def console_url(self):
        return self.job.log_file

    def __repr__(self):
        return self.job.name + " #" + repr(self.build_number)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os
from os.path import join as jp
from jenkinsflow.flow import serial
from .framework import api_select
from .framework.utils import flow_graph_dir

from .json_test import _assert_json

here = os.path.abspath(os.path.dirname(__file__))


with open(jp(here, "just_dump_test_compact.json")) as _jf:
    _compact_json = _jf.read().strip()


def _flow(api, json_dir):
    if json_dir and not os.path.exists(json_dir):
        os.makedirs(json_dir)

    with serial(api, timeout=1, job_name_prefix=api.job_name_prefix, json_dir=json_dir, json_strip_top_level_prefix=True, just_dump=True) as ctrl1:
        ctrl1.invoke('j1')

        with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
            with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                ctrl3a.invoke('j2')
                ctrl3a.invoke_unchecked('j3_unchecked')

            with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                ctrl3b.invoke('j4')

        ctrl1.invoke('j5')

    return ctrl1


def test_just_dump_no_json(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j3_unchecked', max_fails=0, expect_invocations=0, expect_order=None, exec_time=40)
        api.job('j4', max_fails=0, expect_invocations=0, expect_order=None, exec_time=5)
        api.job('j5', max_fails=0, expect_invocations=0, expect_order=None, exec_time=5)

        _flow(api, None)


def test_just_dump_with_json(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        flow_name = api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j3_unchecked', max_fails=0, expect_invocations=0, expect_order=None, exec_time=40)
        api.job('j4', max_fails=0, expect_invocations=0, expect_order=None, exec_time=5)
        api.job('j5', max_fails=0, expect_invocations=0, expect_order=None, exec_time=5)

        fgd = flow_graph_dir(flow_name)
        ctrl1 = _flow(api, fgd)

        # Test json
        json = ctrl1.json(None)
        _assert_json(json, _compact_json, api.api_type)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial
from .framework import api_select


def test_multi_level_mixed(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))
        api.job('job-1', max_fails=0, expect_invocations=1, expect_order=1, params=_params)
        api.job('job-2', max_fails=0, expect_invocations=1, expect_order=2, params=_params, serial=True)
        api.job('job-3', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-4', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-5', max_fails=0, expect_invocations=1, expect_order=3, params=_params)
        api.job('job-6', max_fails=0, expect_invocations=1, expect_order=4, params=_params)
        api.job('job-7', max_fails=0, expect_invocations=1, expect_order=5, params=_params, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('job-1', password='a', s1='b')
            ctrl1.invoke('job-2', password='a', s1='b')

            with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                    ctrl3a.invoke('job-3', password='a', s1='b')
                    ctrl3a.invoke('job-6', password='a', s1='b')

                with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                    ctrl3b.invoke('job-4', password='a', s1='b')
                    ctrl3b.invoke('job-5', password='a', s1='b')

            ctrl1.invoke('job-7', password='a', s1='b')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import serial, MessageRedefinedException
from .framework import api_select
from .framework.utils import lines_in, result_msg


def test_messages(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=1, serial=True)
        api.job('j21', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=2, serial=True)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=3, serial=True)
        api.job('j22', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=4)
        api.job('j23', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=2.00, expect_order=4)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial() as sctrl2:
                sctrl2.message("**** Message 1 ****")
                sctrl2.invoke('j21')
            ctrl1.invoke('j12')
            with ctrl1.parallel() as pctrl:
                pctrl.message("==== Message 2 ====")
                pctrl.invoke('j22')
                pctrl.invoke('j23')

        sout, _ = capsys.readouterr()
        print(sout)
        assert lines_in(
            api_type, sout,
            "^--- Starting flow ---",
            "^Flow Invocation (1/1,1/1): ['jenkinsflow_test__messages__j11', ['jenkinsflow_test__messages__j21'], 'jenkinsflow_test__messages__j12', (",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages__j11",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages__j11"),
            "^**** Message 1 ****",
            "^Flow Invocation (1/1,1/1): ['jenkinsflow_test__messages__j21']",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages__j21",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages__j21"),
            "^Flow SUCCESS ['jenkinsflow_test__messages__j21'] after: ",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages__j12",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages__j12"),
            "^==== Message 2 ====",
            "^Flow Invocation (1/1,1/1): ('jenkinsflow_test__messages__j22', 'jenkinsflow_test__messages__j23')",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages__j22",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages__j23",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages__j22"),
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages__j23"),
            "^Flow SUCCESS ('jenkinsflow_test__messages__j22', 'jenkinsflow_test__messages__j23') after: ",
            "^Flow SUCCESS ['jenkinsflow_test__messages__j11', ['jenkinsflow_test__messages__j21'], 'jenkinsflow_test__messages__j12', ('jenkinsflow_test__messages__j22', 'jenkinsflow_test__messages__j23')]",
        )


def test_messages_on_job(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j21', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=1, serial=True)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=2, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            with ctrl1.serial() as sctrl2:
                with sctrl2.invoke('j21') as j21:
                    j21.message("*** Calling j21 ***")
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()
        print(sout)
        assert lines_in(
            api_type, sout,
            "^--- Starting flow ---",
            "^Flow Invocation (1/1,1/1): [['jenkinsflow_test__messages_on_job__j21'], 'jenkinsflow_test__messages_on_job__j12'",
            "^Flow Invocation (1/1,1/1): ['jenkinsflow_test__messages_on_job__j21']",
            "^*** Calling j21 ***",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages_on_job__j21",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages_on_job__j21"),
            "^Flow SUCCESS ['jenkinsflow_test__messages_on_job__j21'] after: ",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__messages_on_job__j12",
            "^SUCCESS: " + result_msg(api, "jenkinsflow_test__messages_on_job__j12"),
        )


def test_messages_redefined(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=0, invocation_delay=1.0, expect_order=None)
        api.job('j21', max_fails=0, expect_invocations=0, invocation_delay=1.0, expect_order=None)

        with raises(MessageRedefinedException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j11')
                with ctrl1.serial() as sctrl2:
                    sctrl2.message("**** Message 1 ****")
                    sctrl2.invoke('j21')
                    sctrl2.message("New message")
                ctrl1.invoke('j12')

        assert str(exinfo.value) == "Existing message: '**** Message 1 ****', new message: 'New message'"



# Copyright (c) 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, multiprocessing
import time

import bottle
import requests
import pytest
from pytest import raises

from jenkinsflow import jenkins_api
from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in


here = os.path.abspath(os.path.dirname(__file__))


@bottle.route('/')
def _index():
    return bottle.static_file('which_ci_server.html', root=here)


@bottle.route('/api/json')
def _api():
    return bottle.static_file('which_ci_server.html', root=here)


_host = 'localhost'
_port = 8082


def _server():
    bottle.run(host=_host, port=_port, debug=True)


@pytest.mark.apis(ApiType.JENKINS)
def test_which_ci_server_not_ci(api_type):
    proc = None
    try:
        with api_select.api(__file__, api_type) as api:
            proc = multiprocessing.Process(target=_server)
            proc.start()

            with raises(Exception) as exinfo:
                for _ in range(0, 10):
                    ex = None
                    try:
                        jenkins_api.Jenkins("http://" + _host + ':' + repr(_port), "dummy").poll()
                    except requests.exceptions.ConnectionError as ex:
                        # Wait for bottle to start
                        print(ex)
                        time.sleep(0.1)

            assert lines_in(
                api_type, str(exinfo.value),
                 "Not connected to Jenkins or Hudson (expected X-Jenkins or X-Hudson header, got: "
            )

    finally:
        if proc:
            proc.terminate()
